@@ -47,6 +47,10 @@
  *  Some more i18n.
  * Sun Jul 18 03:19:42 MEST 1999 <aeb@cwi.nl>
  *  Terabyte-sized disks.
+ * Sat Jun 30 05:23:19 EST 2001 <nathans@sgi.com>
+ *  XFS label recognition.
+ * Thu Nov 22 15:42:56 CET 2001 <flavio.stanchina@tin.it>
+ *  ext3 and ReiserFS recognition.
  *
  ****************************************************************************/
 
@@ -73,10 +77,9 @@
 #include <sys/stat.h>
 #include <sys/ioctl.h>
 #include <linux/types.h>
-#include <linux/hdreg.h>
-#include <linux/fs.h>		/* for BLKRRPART, BLKGETSIZE */
 
 #include "nls.h"
+#include "xstrncpy.h"
 #include "common.h"
 
 #if defined(__GNUC__) || defined(HAS_LONG_LONG)
@@ -88,7 +91,6 @@ typedef long      ext2_loff_t;
 extern ext2_loff_t ext2_llseek(unsigned int fd, ext2_loff_t offset,
 			       unsigned int origin);
 
-#include "../version.h"
 #define VERSION UTIL_LINUX_VERSION
 
 #define DEFAULT_DEVICE "/dev/hda"
@@ -207,7 +209,7 @@ int kern_heads = 0, kern_sectors = 0;
 int pt_heads = 0, pt_sectors = 0;
 
 
-void
+static void
 set_hsc0(unsigned char *h, unsigned char *s, int *c, int sector) {
 	if (sector >= 1024*cylinder_size)
 		sector = 1024*cylinder_size - 1;
@@ -218,7 +220,7 @@ set_hsc0(unsigned char *h, unsigned char *s, int *c, int sector) {
 	*c = sector;
 }
 
-void
+static void
 set_hsc(unsigned char *h, unsigned char *s, unsigned char *c, int sector) {
 	int cc;
 
@@ -227,12 +229,12 @@ set_hsc(unsigned char *h, unsigned char *s, unsigned char *c, int sector) {
 	*s |= (cc >> 2) & 0xC0;
 }
 
-void
+static void
 set_hsc_begin(struct partition *p, int sector) {
 	set_hsc(& p->head, & p->sector, & p->cyl, sector);
 }
 
-void
+static void
 set_hsc_end(struct partition *p, int sector) {
 	set_hsc(& p->end_head, & p->end_sector, & p->end_cyl, sector);
 }
@@ -247,7 +249,7 @@ set_hsc_end(struct partition *p, int sector) {
 
 /* start_sect and nr_sects are stored little endian on all machines */
 /* moreover, they are not aligned correctly */
-void
+static void
 store4_little_endian(unsigned char *cp, unsigned int val) {
 	cp[0] = (val & 0xff);
 	cp[1] = ((val >> 8) & 0xff);
@@ -255,28 +257,28 @@ store4_little_endian(unsigned char *cp, unsigned int val) {
 	cp[3] = ((val >> 24) & 0xff);
 }
 
-unsigned int
+static unsigned int
 read4_little_endian(unsigned char *cp) {
 	return (uint)(cp[0]) + ((uint)(cp[1]) << 8)
 		+ ((uint)(cp[2]) << 16) + ((uint)(cp[3]) << 24);
 }
 
-void
+static void
 set_start_sect(struct partition *p, unsigned int start_sect) {
 	store4_little_endian(p->start4, start_sect);
 }
 
-unsigned int
+static unsigned int
 get_start_sect(struct partition *p) {
 	return read4_little_endian(p->start4);
 }
 
-void
+static void
 set_nr_sects(struct partition *p, unsigned int nr_sects) {
 	store4_little_endian(p->size4, nr_sects);
 }
 
-unsigned int
+static unsigned int
 get_nr_sects(struct partition *p) {
 	return read4_little_endian(p->size4);
 }
@@ -354,11 +356,11 @@ int LABEL_START = 54;
 int SIZE_START = 70;
 int COMMAND_LINE_X = 5;
 
-void die_x(int ret);
-void draw_screen(void);
+static void die_x(int ret);
+static void draw_screen(void);
 
 /* Guaranteed alloc */
-void *
+static void *
 xmalloc (size_t size) {
      void *t;
 
@@ -374,14 +376,14 @@ xmalloc (size_t size) {
 }
 
 /* Some libc's have their own basename() */
-char *my_basename(char *devname)
-{
+static char *
+my_basename(char *devname) {
     char *s = rindex(devname, '/');
     return s ? s+1 : devname;
 }
 
-char *partition_type_name(unsigned char type)
-{
+static char *
+partition_type_name(unsigned char type) {
     struct systypes *s = i386_sys_types;
 
     while(s->name && s->type != type)
@@ -389,8 +391,8 @@ char *partition_type_name(unsigned char type)
     return s->name;
 }
 
-char *partition_type_text(int i)
-{
+static char *
+partition_type_text(int i) {
     if (p_info[i].id == UNUSABLE)
 	 return _("Unusable");
     else if (p_info[i].id == FREE_SPACE)
@@ -398,6 +400,12 @@ char *partition_type_text(int i)
     else if (p_info[i].id == LINUX) {
 	 if (!strcmp(p_info[i].fstype, "ext2"))
 	      return _("Linux ext2");
+	 else if (!strcmp(p_info[i].fstype, "ext3"))
+	      return _("Linux ext3");
+	 else if (!strcmp(p_info[i].fstype, "xfs"))
+	      return _("Linux XFS");
+	 else if (!strcmp(p_info[i].fstype, "reiserfs"))
+	      return _("Linux ReiserFS");
 	 else
 	      return _("Linux");
     } else if (p_info[i].id == OS2_OR_NTFS) {
@@ -413,8 +421,8 @@ char *partition_type_text(int i)
 	 return partition_type_name(p_info[i].id);
 }
 
-void fdexit(int ret)
-{
+static void
+fdexit(int ret) {
     if (opened)
 	close(fd);
 
@@ -431,8 +439,8 @@ void fdexit(int ret)
     exit(ret);
 }
 
-int get_string(char *str, int len, char *def)
-{
+static int
+get_string(char *str, int len, char *def) {
     char c;
     int i = 0;
     int x, y;
@@ -490,8 +498,8 @@ int get_string(char *str, int len, char *def)
 	return i;
 }
 
-void clear_warning(void)
-{
+static void
+clear_warning(void) {
     int i;
 
     if (!curses_started || !warning_last_time)
@@ -504,8 +512,8 @@ void clear_warning(void)
     warning_last_time = FALSE;
 }
 
-void print_warning(char *s)
-{
+static void
+print_warning(char *s) {
     if (!curses_started) {
 	 fprintf(stderr, "%s\n", s);
     } else {
@@ -516,36 +524,39 @@ void print_warning(char *s)
     }
 }
 
-void fatal(char *s, int ret)
-{
-    char *err = _("FATAL ERROR");
+static void
+fatal(char *s, int ret) {
+    char *err1 = _("FATAL ERROR");
+    char *err2 = _("Press any key to exit cfdisk");
 
     if (curses_started) {
-	 char *str = xmalloc(strlen(s) + strlen(err) + 10);
-	 /* snprintf does not compile on all libc's */
-	 sprintf(str, "%s: %s", err, s);
+	 char *str = xmalloc(strlen(s) + strlen(err1) + strlen(err2) + 10);
+
+	 sprintf(str, "%s: %s", err1, s);
 	 if (strlen(str) > COLS)
-		 str[COLS] = 0;
+	     str[COLS] = 0;
 	 mvaddstr(WARNING_START, (COLS-strlen(str))/2, str);
-	 sprintf(str, _("Press any key to exit cfdisk"));
+	 sprintf(str, "%s", err2);
+	 if (strlen(str) > COLS)
+	     str[COLS] = 0;
 	 mvaddstr(WARNING_START+1, (COLS-strlen(str))/2, str);
 	 putchar(BELL); /* CTRL-G */
 	 refresh();
 	 (void)getch();
 	 die_x(ret);
     } else {
-	 fprintf(stderr, "%s: %s\n", err, s);
+	 fprintf(stderr, "%s: %s\n", err1, s);
 	 exit(ret);
     }
 }
 
-void die(int dummy)
-{
+static void
+die(int dummy) {
     die_x(0);
 }
 
-void die_x(int ret)
-{
+static void
+die_x(int ret) {
     signal(SIGINT, old_SIGINT);
     signal(SIGTERM, old_SIGTERM);
 #ifdef SLCURSES
@@ -560,23 +571,24 @@ void die_x(int ret)
     fdexit(ret);
 }
 
-void read_sector(char *buffer, int sect_num)
-{
+static void
+read_sector(char *buffer, int sect_num) {
     if (ext2_llseek(fd, ((ext2_loff_t) sect_num)*SECTOR_SIZE, SEEK_SET) < 0)
 	fatal(_("Cannot seek on disk drive"), 2);
     if (read(fd, buffer, SECTOR_SIZE) != SECTOR_SIZE)
 	fatal(_("Cannot read disk drive"), 2);
 }
 
-void write_sector(char *buffer, int sect_num)
-{
+static void
+write_sector(char *buffer, int sect_num) {
     if (ext2_llseek(fd, ((ext2_loff_t) sect_num)*SECTOR_SIZE, SEEK_SET) < 0)
 	fatal(_("Cannot seek on disk drive"), 2);
     if (write(fd, buffer, SECTOR_SIZE) != SECTOR_SIZE)
 	fatal(_("Cannot write disk drive"), 2);
 }
 
-void dos_copy_to_info(char *to, int tosz, char *from, int fromsz) {
+static void
+dos_copy_to_info(char *to, int tosz, char *from, int fromsz) {
      int i;
 
      for(i=0; i<tosz && i<fromsz && isascii(from[i]); i++)
@@ -584,62 +596,122 @@ void dos_copy_to_info(char *to, int tosz, char *from, int fromsz) {
      to[i] = 0;
 }
 
-void get_dos_label(int i)
-{
-    char sector[128];
+static void
+get_dos_label(int i) {
+	char sector[128];
 #define DOS_OSTYPE_OFFSET 3
 #define DOS_LABEL_OFFSET 43
 #define DOS_FSTYPE_OFFSET 54
 #define DOS_OSTYPE_SZ 8
 #define DOS_LABEL_SZ 11
 #define DOS_FSTYPE_SZ 8
-    ext2_loff_t offset;
+	ext2_loff_t offset;
 
-    offset = ((ext2_loff_t) p_info[i].first_sector + p_info[i].offset)
-	 * SECTOR_SIZE;
-    if (ext2_llseek(fd, offset, SEEK_SET) == offset
-	&& read(fd, &sector, sizeof(sector)) == sizeof(sector)) {
-	 dos_copy_to_info(p_info[i].ostype, OSTYPESZ,
-			  sector+DOS_OSTYPE_OFFSET, DOS_OSTYPE_SZ);
-	 dos_copy_to_info(p_info[i].volume_label, LABELSZ,
-			  sector+DOS_LABEL_OFFSET, DOS_LABEL_SZ);
-	 dos_copy_to_info(p_info[i].fstype, FSTYPESZ,
-			  sector+DOS_FSTYPE_OFFSET, DOS_FSTYPE_SZ);
-    }
+	offset = ((ext2_loff_t) p_info[i].first_sector + p_info[i].offset)
+		* SECTOR_SIZE;
+	if (ext2_llseek(fd, offset, SEEK_SET) == offset
+	    && read(fd, &sector, sizeof(sector)) == sizeof(sector)) {
+		dos_copy_to_info(p_info[i].ostype, OSTYPESZ,
+				 sector+DOS_OSTYPE_OFFSET, DOS_OSTYPE_SZ);
+		dos_copy_to_info(p_info[i].volume_label, LABELSZ,
+				 sector+DOS_LABEL_OFFSET, DOS_LABEL_SZ);
+		dos_copy_to_info(p_info[i].fstype, FSTYPESZ,
+				 sector+DOS_FSTYPE_OFFSET, DOS_FSTYPE_SZ);
+	}
 }
 
-void get_ext2_label(int i)
-{
-#define EXT2_SUPER_MAGIC 0xEF53
+#define REISERFS_SUPER_MAGIC_STRING "ReIsErFs"
+#define REISER2FS_SUPER_MAGIC_STRING "ReIsEr2Fs"
+struct reiserfs_super_block {
+	char s_dummy0[ 52];
+	char s_magic [ 12];
+	char s_dummy1[140];
+};
+
+static int
+is_reiserfs_magic_string(const struct reiserfs_super_block *rs) {
+	return (!strncmp(rs->s_magic, REISERFS_SUPER_MAGIC_STRING,
+			 strlen(REISERFS_SUPER_MAGIC_STRING)) ||
+		!strncmp(rs->s_magic, REISER2FS_SUPER_MAGIC_STRING,
+			 strlen(REISER2FS_SUPER_MAGIC_STRING)));
+}
+
+static void
+get_linux_label(int i) {
+
 #define EXT2LABELSZ 16
-    struct ext2_super_block {
-	char  s_dummy0[56];
-	unsigned char  s_magic[2];
-	char  s_dummy1[62];
-	char  s_volume_name[EXT2LABELSZ];
-	char  s_last_mounted[64];
-	char  s_dummy2[824];
-    } sb;
-    char *label = sb.s_volume_name;
-    ext2_loff_t offset;
-    int j;
+#define EXT2_SUPER_MAGIC 0xEF53
+#define EXT3_FEATURE_COMPAT_HAS_JOURNAL 0x0004
+	struct ext2_super_block {
+		char  s_dummy0[56];
+		unsigned char  s_magic[2];
+		char  s_dummy1[34];
+		unsigned char  s_feature_compat[4];
+		char  s_dummy2[24];
+		char  s_volume_name[EXT2LABELSZ];
+		char  s_last_mounted[64];
+		char  s_dummy3[824];
+	} e2fsb;
 
-    offset = ((ext2_loff_t) p_info[i].first_sector + p_info[i].offset)
-	 * SECTOR_SIZE + 1024;
-    if (ext2_llseek(fd, offset, SEEK_SET) == offset
-	&& read(fd, &sb, sizeof(sb)) == sizeof(sb)
-	&& sb.s_magic[0] + 256*sb.s_magic[1] == EXT2_SUPER_MAGIC) {
-	 for(j=0; j<EXT2LABELSZ; j++)
-	      if(!isprint(label[j]))
-		   label[j] = 0;
-	 label[EXT2LABELSZ] = 0;
-	 strncpy(p_info[i].volume_label, label, LABELSZ);
-	 strncpy(p_info[i].fstype, "ext2", FSTYPESZ);
-    }
+#define REISERFS_DISK_OFFSET_IN_BYTES (64 * 1024)
+	struct reiserfs_super_block reiserfsb;
+
+#define XFS_SUPER_MAGIC "XFSB"
+#define XFSLABELSZ 12
+	struct xfs_super_block {
+		unsigned char   s_magic[4];
+		unsigned char   s_dummy0[104];
+		unsigned char   s_fname[XFSLABELSZ];
+		unsigned char   s_dummy1[904];
+	} xfsb;
+
+	char *label;
+	ext2_loff_t offset;
+	int j;
+
+	offset = ((ext2_loff_t) p_info[i].first_sector + p_info[i].offset)
+		* SECTOR_SIZE + 1024;
+	if (ext2_llseek(fd, offset, SEEK_SET) == offset
+	    && read(fd, &e2fsb, sizeof(e2fsb)) == sizeof(e2fsb)
+	    && e2fsb.s_magic[0] + (e2fsb.s_magic[1]<<8) == EXT2_SUPER_MAGIC) {
+		label = e2fsb.s_volume_name;
+		for(j=0; j<EXT2LABELSZ && j<LABELSZ && isprint(label[j]); j++)
+			p_info[i].volume_label[j] = label[j];
+		p_info[i].volume_label[j] = 0;
+		/* ext2 or ext3? */
+		if (e2fsb.s_feature_compat[0]&EXT3_FEATURE_COMPAT_HAS_JOURNAL)
+			strncpy(p_info[i].fstype, "ext3", FSTYPESZ);
+		else
+			strncpy(p_info[i].fstype, "ext2", FSTYPESZ);
+		return;
+	}
+
+	offset = ((ext2_loff_t) p_info[i].first_sector + p_info[i].offset)
+		* SECTOR_SIZE + 0;
+	if (ext2_llseek(fd, offset, SEEK_SET) == offset
+	    && read(fd, &xfsb, sizeof(xfsb)) == sizeof(xfsb)
+	    && !strcmp(xfsb.s_magic, XFS_SUPER_MAGIC)) {
+		label = xfsb.s_fname;
+		for(j=0; j<XFSLABELSZ && j<LABELSZ && isprint(label[j]); j++)
+			p_info[i].volume_label[j] = label[j];
+		p_info[i].volume_label[j] = 0;
+		strncpy(p_info[i].fstype, "xfs", FSTYPESZ);
+		return;
+	}
+
+	/* reiserfs? */
+	offset = ((ext2_loff_t) p_info[i].first_sector + p_info[i].offset)
+		* SECTOR_SIZE + REISERFS_DISK_OFFSET_IN_BYTES;
+	if (ext2_llseek(fd, offset, SEEK_SET) == offset
+	    && read(fd, &reiserfsb, 1024) == 1024
+	    && is_reiserfs_magic_string(&reiserfsb)) {
+		strncpy(p_info[i].fstype, "reiserfs", FSTYPESZ);
+		return;
+	}
 }
 
-void check_part_info(void)
-{
+static void
+check_part_info(void) {
     int i, pri = 0, log = 0;
 
     for (i = 0; i < num_parts; i++)
@@ -719,8 +791,8 @@ void check_part_info(void)
     }
 }
 
-void remove_part(int i)
-{
+static void
+remove_part(int i) {
     int p;
 
     for (p = i; p < num_parts; p++)
@@ -731,8 +803,8 @@ void remove_part(int i)
 	cur_part--;
 }
 
-void insert_empty_part(int i, int first, int last)
-{
+static void
+insert_empty_part(int i, int first, int last) {
     int p;
 
     for (p = num_parts; p > i; p--)
@@ -751,8 +823,8 @@ void insert_empty_part(int i, int first, int last)
     num_parts++;
 }
 
-void del_part(int i)
-{
+static void
+del_part(int i) {
     int num = p_info[i].num;
 
     if (i > 0 && (p_info[i-1].id == FREE_SPACE ||
@@ -806,9 +878,9 @@ void del_part(int i)
     check_part_info();
 }
 
-int add_part(int num, int id, int flags, int first, int last, int offset,
-	     int want_label, char **errmsg)
-{
+static int
+add_part(int num, int id, int flags, int first, int last, int offset,
+	 int want_label, char **errmsg) {
     int i, pri = 0, log = 0;
 
     if (num_parts == MAXIMUM_PARTS) {
@@ -945,7 +1017,7 @@ int add_part(int num, int id, int flags, int first, int last, int offset,
 	 if (may_have_dos_label(id))
 	      get_dos_label(i);
 	 else if (id == LINUX)
-	      get_ext2_label(i);
+	      get_linux_label(i);
     }
 
     check_part_info();
@@ -953,8 +1025,8 @@ int add_part(int num, int id, int flags, int first, int last, int offset,
     return 0;
 }
 
-int find_primary(void)
-{
+static int
+find_primary(void) {
     int num = 0, cur = 0;
 
     while (cur < num_parts && IS_PRIMARY(num))
@@ -971,8 +1043,8 @@ int find_primary(void)
 	return num;
 }
 
-int find_logical(int i)
-{
+static int
+find_logical(int i) {
     int num = -1;
     int j;
 
@@ -990,8 +1062,8 @@ int find_logical(int i)
     return num;
 }
 
-void inc_logical(int i)
-{
+static void
+inc_logical(int i) {
     int j;
 
     for (j = i; j < num_parts; j++)
@@ -1026,26 +1098,24 @@ struct MenuItem
  * Should not be called directly. Call function menuSelect instead.
  */
 
-int menuUpdate( int y, int x, struct MenuItem *menuItems, int itemLength,
-		char *available, int menuType, int current )
-{
+static int
+menuUpdate( int y, int x, struct MenuItem *menuItems, int itemLength,
+	    char *available, int menuType, int current ) {
     int i, lmargin = x, ymargin = y;
     char *mcd;
 
     /* Print available buttons */
     move( y, x ); clrtoeol();
 
-    for( i = 0; menuItems[i].key; i++ )
-    {
+    for( i = 0; menuItems[i].key; i++ ) {
         char buff[20];
         int lenName;
 	const char *mi;
 
         /* Search next available button */
         while( menuItems[i].key && !strchr(available, menuItems[i].key) )
-        {
             i++;
-        }
+
         if( !menuItems[i].key ) break; /* No more menu items */
 
         /* If selected item is not available and we have bypassed it,
@@ -1067,15 +1137,13 @@ int menuUpdate( int y, int x, struct MenuItem *menuItems, int itemLength,
             print_warning(_("Menu item too long. Menu may look odd."));
 #endif
 	if (lenName >= sizeof(buff)) {	/* truncate ridiculously long string */
-	    strncpy( buff, mi, sizeof(buff)-1);
-	    buff[sizeof(buff)-1] = 0;
-	} else
-        if( menuType & MENU_BUTTON )
-            sprintf( buff, "[%*s%-*s]", (itemLength - lenName) / 2, "",
-                (itemLength - lenName + 1) / 2 + lenName, mi );
-        else
-            sprintf( buff, "%*s%-*s", (itemLength - lenName) / 2, "",
-                (itemLength - lenName + 1) / 2 + lenName, mi );
+	    xstrncpy(buff, mi, sizeof(buff));
+	} else {
+            snprintf(buff, sizeof(buff),
+		     (menuType & MENU_BUTTON) ? "[%*s%-*s]" : "%*s%-*s",
+		     (itemLength - lenName) / 2, "",
+		     (itemLength - lenName + 1) / 2 + lenName, mi);
+	}
         mvaddstr( y, x, buff );
 
         /* Lowlight after selected item */
@@ -1113,9 +1181,9 @@ int menuUpdate( int y, int x, struct MenuItem *menuItems, int itemLength,
 /* This function takes a list of menu items, lets the user choose one *
  * and returns the value keyboard shortcut of the selected menu item  */
 
-int menuSelect( int y, int x, struct MenuItem *menuItems, int itemLength,
-		char *available, int menuType, int menuDefault )
-{
+static int
+menuSelect( int y, int x, struct MenuItem *menuItems, int itemLength,
+	    char *available, int menuType, int menuDefault ) {
     int i, ylast = y, key = 0, current = menuDefault;
 
     if( !( menuType & ( MENU_HORIZ | MENU_VERT ) ) )	
@@ -1269,8 +1337,8 @@ int menuSelect( int y, int x, struct MenuItem *menuItems, int itemLength,
  * and waits for a keypress.                                            *
  * Perhaps calling function menuSelect is a bit overkill but who cares? */
 
-void menuContinue(void)
-{
+static void
+menuContinue(void) {
     static struct MenuItem menuContinueBtn[]=
     {
         { 'c', "", N_("Press a key to continue") },
@@ -1284,8 +1352,8 @@ void menuContinue(void)
 /* Function menuSelect takes way too many parameters  *
  * Luckily, most of time we can do with this function */
 
-int menuSimple(struct MenuItem *menuItems, int menuDefault)
-{
+static int
+menuSimple(struct MenuItem *menuItems, int menuDefault) {
     int i, j, itemLength = 0;
     char available[MENU_MAX_ITEMS];
 
@@ -1302,8 +1370,8 @@ int menuSimple(struct MenuItem *menuItems, int menuDefault)
 
 /* End of command menu support code */
 
-void new_part(int i)
-{
+static void
+new_part(int i) {
     char response[LINE_LENGTH], def[LINE_LENGTH];
     char c;
     int first = p_info[i].first_sector;
@@ -1340,7 +1408,7 @@ void new_part(int i)
     else
 	print_warning(_("!!! Internal error !!!"));
 
-    sprintf(def, "%.2f", num_sects/sectors_per_MB);
+    snprintf(def, sizeof(def), "%.2f", num_sects/sectors_per_MB);
     mvaddstr(COMMAND_LINE_Y, COMMAND_LINE_X, _("Size (in MB): "));
     if ((len = get_string(response, LINE_LENGTH, def)) <= 0 &&
 	len != GS_DEFAULT)
@@ -1412,8 +1480,8 @@ void new_part(int i)
     (void) add_part(num, id, flags, first, last, offset, 0, &errmsg);
 }
 
-void get_kernel_geometry(void)
-{
+static void
+get_kernel_geometry(void) {
 #ifdef HDIO_GETGEO
     struct hd_geometry geometry;
 
@@ -1424,16 +1492,37 @@ void get_kernel_geometry(void)
 #endif
 }
 
-void get_partition_table_geometry(partition_table *bufp)
-{
+static void
+get_partition_table_geometry(partition_table *bufp) {
     struct partition *p;
     int i,h,s,hh,ss;
     int first = TRUE;
     int bad = FALSE;
 
     if (bufp->p.magicflag[0] != PART_TABLE_FLAG0 ||
-	bufp->p.magicflag[1] != PART_TABLE_FLAG1)
-	fatal(_("Bad signature on partition table"), 3);
+	bufp->p.magicflag[1] != PART_TABLE_FLAG1) {
+	    /* Matthew Wilcox: slightly friendlier version of
+	       fatal(_("Bad signature on partition table"), 3);
+	    */
+	    int cont;
+	    mvaddstr(WARNING_START, 0,
+	      _("No partition table or unknown signature on partition table"));
+	    mvaddstr(WARNING_START+1, 0,
+	      _("Do you wish to start with a zero table [y/N] ?"));
+	    putchar(BELL);
+	    refresh();
+	    cont = getch();
+	    if ((cont != 'y') && (cont != 'Y'))
+		    die_x(3);
+	    zero_table = TRUE;
+	    return;
+
+	    /* Oskar Liljeblad suggested:
+	         Bad signature blah blah
+		 If this is a brand new harddrive that has not been partitioned
+		 before, please run cfdisk -z.
+	    */
+    }
 
     hh = ss = 0;
     for (i=0; i<4; i++) {
@@ -1456,8 +1545,8 @@ void get_partition_table_geometry(partition_table *bufp)
     }
 }
 
-void decide_on_geometry(void)
-{
+static void
+decide_on_geometry(void) {
     heads = (user_heads ? user_heads :
 	     pt_heads ? pt_heads :
 	     kern_heads ? kern_heads : 255);
@@ -1474,8 +1563,8 @@ void decide_on_geometry(void)
 	    print_warning(_("You specified more cylinders than fit on disk"));
 }
 
-void clear_p_info(void)
-{
+static void
+clear_p_info(void) {
     num_parts = 1;
     p_info[0].first_sector = 0;
     p_info[0].last_sector = total_size - 1;
@@ -1492,8 +1581,8 @@ void clear_p_info(void)
     ext_info.num = PRIMARY;
 }
 
-void fill_p_info(void)
-{
+static void
+fill_p_info(void) {
     int pn, i, bs, bsz;
     struct partition *p;
     partition_table buffer;
@@ -1549,7 +1638,6 @@ void fill_p_info(void)
 		add_part(i, p->sys_ind, p->boot_ind,
 			 ((bs <= sectors) ? 0 : bs), bs + bsz - 1,
 			 ((bs <= sectors) ? bs : 0), 1, &errmsg)) {
-		    /* avoid snprintf - it does not exist on ancient systems */
 		    char *bad = _("Bad primary partition");
 		    char *msg = (char *) xmalloc(strlen(bad) + strlen(errmsg) + 30);
 		    sprintf(msg, "%s %d: %s", bad, i, errmsg);
@@ -1580,7 +1668,6 @@ void fill_p_info(void)
 			     logical_sectors[logical-1],
 			     logical_sectors[logical-1] + bs + bsz - 1,
 			     bs, 1, &errmsg)) {
-				/* avoid snprintf */
 				char *bad = _("Bad logical partition");
 				char *msg = (char *) xmalloc(strlen(bad) + strlen(errmsg) + 30);
 				sprintf(msg, "%s %d: %s", bad, i-1, errmsg);
@@ -1603,8 +1690,8 @@ void fill_p_info(void)
     }
 }
 
-void fill_part_table(struct partition *p, partition_info *pi)
-{
+static void
+fill_part_table(struct partition *p, partition_info *pi) {
     int begin;
 
     p->boot_ind = pi->flags;
@@ -1619,8 +1706,8 @@ void fill_part_table(struct partition *p, partition_info *pi)
     set_hsc_end(p, pi->last_sector);
 }
 
-void fill_primary_table(partition_table *buffer)
-{
+static void
+fill_primary_table(partition_table *buffer) {
     int i;
 
     /* Zero out existing table */
@@ -1638,8 +1725,8 @@ void fill_primary_table(partition_table *buffer)
     buffer->p.magicflag[1] = PART_TABLE_FLAG1;
 }
 
-void fill_logical_table(partition_table *buffer, partition_info *pi)
-{
+static void
+fill_logical_table(partition_table *buffer, partition_info *pi) {
     struct partition *p;
     int i;
 
@@ -1675,8 +1762,8 @@ void fill_logical_table(partition_table *buffer, partition_info *pi)
     buffer->p.magicflag[1] = PART_TABLE_FLAG1;
 }
 
-void write_part_table(void)
-{
+static void
+write_part_table(void) {
     int i, ct, done = FALSE, len;
     partition_table buffer;
     struct stat s;
@@ -1752,18 +1839,20 @@ void write_part_table(void)
     for (i = 0; i < num_parts; i++)
 	if (IS_PRIMARY(i) && p_info[i].flags == ACTIVE_FLAG)
 	    ct++;
-    if (ct != 1)
-	print_warning(_("Not precisely one primary partition is bootable. DOS MBR cannot boot this."));
+    if (ct == 0)
+	print_warning(_("No primary partitions are marked bootable. DOS MBR cannot boot this."));
+    if (ct > 1)
+	print_warning(_("More than one primary partition is marked bootable. DOS MBR cannot boot this."));
 }
 
-void fp_printf(FILE *fp, char *format, ...)
-{
+static void
+fp_printf(FILE *fp, char *format, ...) {
     va_list args;
     char buf[1024];
     int y, x;
 
     va_start(args, format);
-    vsprintf(buf, format, args);
+    vsnprintf(buf, sizeof(buf), format, args);
     va_end(args);
 
     if (fp == NULL) {
@@ -1781,8 +1870,8 @@ void fp_printf(FILE *fp, char *format, ...)
 }
 
 #define MAX_PER_LINE 16
-void print_file_buffer(FILE *fp, char *buffer)
-{
+static void
+print_file_buffer(FILE *fp, char *buffer) {
     int i,l;
 
     for (i = 0, l = 0; i < SECTOR_SIZE; i++, l++) {
@@ -1799,8 +1888,8 @@ void print_file_buffer(FILE *fp, char *buffer)
     fp_printf(fp, "\n");
 }
 
-void print_raw_table(void)
-{
+static void
+print_raw_table(void) {
     int i, to_file;
     partition_table buffer;
     char fname[LINE_LENGTH];
@@ -1819,7 +1908,8 @@ void print_raw_table(void)
 	if (to_file) {
 	    if ((fp = fopen(fname, "w")) == NULL) {
 		char errstr[LINE_LENGTH];
-		sprintf(errstr, _("Cannot open file '%s'"), fname);
+		snprintf(errstr, sizeof(errstr),
+			 _("Cannot open file '%s'"), fname);
 		print_warning(errstr);
 		return;
 	    }
@@ -1853,8 +1943,8 @@ void print_raw_table(void)
     }
 }
 
-void print_p_info_entry(FILE *fp, partition_info *p)
-{
+static void
+print_p_info_entry(FILE *fp, partition_info *p) {
     int size;
     char part_str[40];
 
@@ -1918,8 +2008,8 @@ void print_p_info_entry(FILE *fp, partition_info *p)
     fp_printf(fp, "\n");
 }
 
-void print_p_info(void)
-{
+static void
+print_p_info(void) {
     char fname[LINE_LENGTH];
     FILE *fp;
     int i, to_file, pext = is_extended(ext_info.id);
@@ -1970,8 +2060,8 @@ void print_p_info(void)
     }
 }
 
-void print_part_entry(FILE *fp, int num, partition_info *pi)
-{
+static void
+print_part_entry(FILE *fp, int num, partition_info *pi) {
     int first = 0, start = 0, end = 0, size = 0;
     unsigned char ss, es, sh, eh;
     int sc, ec;
@@ -2002,8 +2092,8 @@ void print_part_entry(FILE *fp, int num, partition_info *pi)
 }
 
 
-void print_part_table(void)
-{
+static void
+print_part_table(void) {
     int i, j, to_file;
     char fname[LINE_LENGTH];
     FILE *fp;
@@ -2034,6 +2124,7 @@ void print_part_table(void)
 
     fp_printf(fp, _("Partition Table for %s\n"), disk_device);
     fp_printf(fp, "\n");
+    /* Three-line heading. Read "Start Sector" etc vertically. */
     fp_printf(fp, _("         ---Starting---      ----Ending----    Start Number of\n"));
     fp_printf(fp, _(" # Flags Head Sect Cyl   ID  Head Sect Cyl    Sector  Sectors\n"));
     fp_printf(fp, _("-- ----- ---- ---- ---- ---- ---- ---- ---- -------- ---------\n"));
@@ -2063,8 +2154,8 @@ void print_part_table(void)
     }
 }
 
-void print_tables(void)
-{
+static void
+print_tables(void) {
     int done = FALSE;
 
     static struct MenuItem menuFormat[]=
@@ -2097,8 +2188,8 @@ void print_tables(void)
 }
 
 #define END_OF_HELP "EOHS!"
-void display_help()
-{
+static void
+display_help(void) {
     char *help_text[] = {
 	N_("Help Screen for cfdisk"),
 	"",
@@ -2159,8 +2250,8 @@ void display_help()
     menuContinue();
 }
 
-int change_geometry(void)
-{
+static int
+change_geometry(void) {
     int ret_val = FALSE;
     int done = FALSE;
     char def[LINE_LENGTH];
@@ -2278,8 +2369,8 @@ int change_geometry(void)
     return ret_val;
 }
 
-void change_id(int i)
-{
+static void
+change_id(int i) {
     char id[LINE_LENGTH], def[LINE_LENGTH];
     int num_types = 0;
     int num_across, num_down;
@@ -2352,8 +2443,8 @@ void change_id(int i)
 	p_info[i].id = new_id;
 }
 
-void draw_partition(int i)
-{
+static void
+draw_partition(int i) {
     int size, j;
     int y = i + DISK_TABLE_START + 2 - (cur_part/NUM_ON_SCREEN)*NUM_ON_SCREEN;
     char *t;
@@ -2366,8 +2457,13 @@ void draw_partition(int i)
     }
 
     if (p_info[i].id > 0) {
+	char *dbn = my_basename(disk_device);
+	int l = strlen(dbn);
+	int digit_last = isdigit(dbn[l-1]);
+
 	mvprintw(y, NAME_START,
-		 "%s%d", my_basename(disk_device), p_info[i].num+1);
+		 "%s%s%d", dbn, (digit_last ? "p" : ""),
+		 p_info[i].num+1);
 	if (p_info[i].flags) {
 	    if (p_info[i].flags == ACTIVE_FLAG)
 		mvaddstr(y, FLAGS_START, _("Boot"));
@@ -2425,8 +2521,8 @@ void draw_partition(int i)
 	mvprintw(y, COLUMNS-1, "*");
 }
 
-void init_const(void)
-{
+static void
+init_const(void) {
     if (!defined) {
 	NAME_START = (((float)NAME_START)/COLUMNS)*COLS;
 	FLAGS_START = (((float)FLAGS_START)/COLUMNS)*COLS;
@@ -2447,8 +2543,8 @@ void init_const(void)
     }
 }
 
-void draw_screen(void)
-{
+static void
+draw_screen(void) {
     int i;
     char *line;
 
@@ -2472,8 +2568,16 @@ void draw_screen(void)
     mvaddstr(HEADER_START, (COLS-strlen(line))/2, line);
     sprintf(line, _("Disk Drive: %s"), disk_device);
     mvaddstr(HEADER_START+2, (COLS-strlen(line))/2, line);
-    { long long bytes = actual_size*(long long) SECTOR_SIZE;
-    sprintf(line, _("Size: %lld bytes"), bytes); }
+    {
+	    long long bytes = actual_size*(long long) SECTOR_SIZE;
+	    long megabytes = bytes/1000000;
+	    if (megabytes < 10000)
+		    sprintf(line, _("Size: %lld bytes, %ld MB"),
+			    bytes, megabytes);
+	    else
+		    sprintf(line, _("Size: %lld bytes, %ld.%ld GB"),
+			    bytes, megabytes/1000, (megabytes/100)%10);
+    }
     mvaddstr(HEADER_START+3, (COLS-strlen(line))/2, line);
     sprintf(line, _("Heads: %d   Sectors per Track: %d   Cylinders: %d"),
 	    heads, sectors, cylinders);
@@ -2510,8 +2614,8 @@ void draw_screen(void)
     free(line);
 }
 
-int draw_cursor(int move)
-{
+static int
+draw_cursor(int move) {
     if (move != 0 && (cur_part + move < 0 || cur_part + move >= num_parts))
 	return -1;
 
@@ -2539,8 +2643,8 @@ int draw_cursor(int move)
     return 0;
 }
 
-void do_curses_fdisk(void)
-{
+static void
+do_curses_fdisk(void) {
     int done = FALSE;
     char command;
 
@@ -2709,13 +2813,13 @@ void do_curses_fdisk(void)
     die_x(0);
 }
 
-void copyright(void)
-{
-    fprintf(stderr, _("Copyright (C) 1994-1999 Kevin E. Martin & aeb\n"));
+static void
+copyright(void) {
+    fprintf(stderr, _("Copyright (C) 1994-2000 Kevin E. Martin & aeb\n"));
 }
 
-void usage(char *prog_name)
-{
+static void
+usage(char *prog_name) {
     /* Unfortunately, xgettext does not handle multi-line strings */
     /* so, let's use explicit \n's instead */
     fprintf(stderr, _("\n"
@@ -2747,7 +2851,7 @@ main(int argc, char **argv)
     bindtextdomain(PACKAGE, LOCALEDIR);
     textdomain(PACKAGE);
 
-    while ((c = getopt(argc, argv, "ac:gh:s:vzP:")) != EOF)
+    while ((c = getopt(argc, argv, "ac:gh:s:vzP:")) != -1)
 	switch (c) {
 	case 'a':
 	    arrow_cursor = TRUE;
