@@ -8,26 +8,13 @@
  * (at your option) any later version.
  *
  * For detailed old history, see older versions.
- * Contributions before 2000 by faith@cs.unc.edu, Michael Bischoff,
+ * Contributions before 2001 by faith@cs.unc.edu, Michael Bischoff,
  * LeBlanc@mcc.ac.uk, martin@cs.unc.edu, leisner@sdsp.mc.xerox.com,
  * esr@snark.thyrsus.com, aeb@cwi.nl, quinlan@yggdrasil.com,
  * fasten@cs.bonn.edu, orschaer@cip.informatik.uni-erlangen.de,
  * jj@sunsite.mff.cuni.cz, fasten@shw.com, ANeuper@GUUG.de,
- * kgw@suse.de.
- * 
- * Modified, Sun Feb 20 2000, kalium@gmx.de
- * Added fix operation allowing to reorder primary/extended partition
- * entries within the partition table. Some programs or OSes have
- * problems using a partition table with entries not ordered
- * according to their positions on disk.
- * Munged this patch to also make it work for logical partitions.
- * aeb, 2000-02-20.
- *
- * Wed Mar 1 14:34:53 EST 2000 David Huggins-Daines <dhuggins@linuxcare.com>
- * Better support for OSF/1 disklabels on Alpha.
- *
- * 2000-04-06, Michal Jaegermann (michal@ellpspace.math.ualberta.ca)
- * fixed and added some alpha stuff.
+ * kgw@suse.de, kalium@gmx.de, dhuggins@linuxcare.com,
+ * michal@ellpspace.math.ualberta.ca and probably others.
  */
 
 
@@ -42,8 +29,6 @@
 #include <getopt.h>
 #include <sys/stat.h>
 
-#include <linux/hdreg.h>       /* for HDIO_GETGEO */
-
 #include "nls.h"
 #include "common.h"
 #include "fdisk.h"
@@ -66,7 +51,7 @@ static void delete_partition(int i);
 			})
 
 
-#define LINE_LENGTH	80
+#define LINE_LENGTH	800
 #define pt_offset(b, n)	((struct partition *)((b) + 0x1be + \
 				(n) * sizeof(struct partition)))
 #define sector(s)	((s) & 0x3f)
@@ -170,9 +155,12 @@ int	fd,				/* the disk */
 	listing = 0,			/* no aborts for fdisk -l */
 	nowarn = 0,			/* no warnings for fdisk -l/-s */
 	dos_compatible_flag = ~0,
+	dos_changed = 0,
 	partitions = 4;			/* maximum partition + 1 */
 
 uint	user_cylinders, user_heads, user_sectors;
+uint	pt_heads, pt_sectors;
+uint	kern_heads, kern_sectors;
 
 uint	heads,
 	sectors,
@@ -189,6 +177,7 @@ int     sun_label = 0;                  /* looking at sun disklabel */
 int	sgi_label = 0;			/* looking at sgi disklabel */
 int	aix_label = 0;			/* looking at aix disklabel */
 int	osf_label = 0;			/* looking at osf disklabel */
+
 jmp_buf listingbuf;
 
 void fatal(enum failure why) {
@@ -209,7 +198,7 @@ void fatal(enum failure why) {
 "Here DISK is something like /dev/hdb or /dev/sda\n"
 "and PARTITION is something like /dev/hda7\n"
 "-u: give Start and End in sector (instead of cylinder) units\n"
-"-b 2048: (for certain MO drives) use 2048-byte sectors\n");
+"-b 2048: (for certain MO disks) use 2048-byte sectors\n");
 			break;
 		case usage2:
 		  /* msg in cases where fdisk used to probe */
@@ -222,25 +211,31 @@ void fatal(enum failure why) {
 "  ...\n");
 			break;
 		case unable_to_open:
-			sprintf(error, _("Unable to open %s\n"), disk_device);
+			snprintf(error, sizeof(error),
+				 _("Unable to open %s\n"), disk_device);
 			break;
 		case unable_to_read:
-			sprintf(error, _("Unable to read %s\n"), disk_device);
+			snprintf(error, sizeof(error),
+				 _("Unable to read %s\n"), disk_device);
 			break;
 		case unable_to_seek:
-			sprintf(error, _("Unable to seek on %s\n"),disk_device);
+			snprintf(error, sizeof(error),
+				_("Unable to seek on %s\n"),disk_device);
 			break;
 		case unable_to_write:
-			sprintf(error, _("Unable to write %s\n"), disk_device);
+			snprintf(error, sizeof(error),
+				_("Unable to write %s\n"), disk_device);
 			break;
 		case ioctl_error:
-			sprintf(error, _("BLKGETSIZE ioctl failed on %s\n"),
+			snprintf(error, sizeof(error),
+				 _("BLKGETSIZE ioctl failed on %s\n"),
 				disk_device);
 			break;
 		case out_of_memory:
 			message = _("Unable to allocate any more memory\n");
 			break;
-		default: message = _("Fatal error\n");
+		default:
+			message = _("Fatal error\n");
 	}
 
 	fputc('\n', stderr);
@@ -282,7 +277,7 @@ read_pte(int fd, int pno, uint offset) {
 	pe->changed = 0;
 	pe->part_table = pe->ext_pointer = NULL;
 }
-	
+
 static unsigned int
 get_partition_start(struct pte *pe) {
 	return pe->offset + get_start_sect(pe->part_table);
@@ -306,6 +301,21 @@ set_changed(int i) {
 	ptes[i].changed = 1;
 }
 
+/*
+ * Avoid warning about DOS partitions when no DOS partition was changed.
+ * Here a heuristic "is probably dos partition".
+ * We might also do the opposite and warn in all cases except
+ * for "is probably nondos partition".
+ */
+static int
+is_dos_partition(int t) {
+	return (t == 1 || t == 4 || t == 6 ||
+		t == 0x0b || t == 0x0c || t == 0x0e ||
+		t == 0x11 || t == 0x12 || t == 0x14 || t == 0x16 ||
+		t == 0x1b || t == 0x1c || t == 0x1e || t == 0x24 ||
+		t == 0xc1 || t == 0xc4 || t == 0xc6);
+}
+
 static void
 menu(void) {
 	if (sun_label) {
@@ -399,7 +409,7 @@ xmenu(void) {
 	   puts(_("   c   change number of cylinders"));
 	   puts(_("   d   print the raw data in the partition table"));
 	   puts(_("   e   list extended partitions"));		/* !sun */
-	   puts(_("   g   create an IRIX partition table"));	/* sgi */
+	   puts(_("   g   create an IRIX (SGI) partition table"));/* sgi */
 	   puts(_("   h   change number of heads"));
 	   puts(_("   m   print this menu"));
 	   puts(_("   p   print the partition table"));
@@ -415,7 +425,7 @@ xmenu(void) {
 	   puts(_("   c   change number of cylinders"));
 	   puts(_("   d   print the raw data in the partition table"));
 	   puts(_("   e   list extended partitions"));		/* !sun */
-	   puts(_("   g   create an IRIX partition table"));	/* sgi */
+	   puts(_("   g   create an IRIX (SGI) partition table"));/* sgi */
 	   puts(_("   h   change number of heads"));
 	   puts(_("   m   print this menu"));
 	   puts(_("   p   print the partition table"));
@@ -432,7 +442,7 @@ xmenu(void) {
 	   puts(_("   d   print the raw data in the partition table"));
 	   puts(_("   e   list extended partitions"));		/* !sun */
 	   puts(_("   f   fix partition order"));		/* !sun, !aix, !sgi */
-	   puts(_("   g   create an IRIX partition table"));	/* sgi */
+	   puts(_("   g   create an IRIX (SGI) partition table"));/* sgi */
 	   puts(_("   h   change number of heads"));
 	   puts(_("   m   print this menu"));
 	   puts(_("   p   print the partition table"));
@@ -448,14 +458,16 @@ static int
 get_sysid(int i) {
 	return (
 		sun_label ? sunlabel->infos[i].id :
-		sgi_label ? sgi_get_sysid(i) : ptes[i].part_table->sys_ind);
+		sgi_label ? sgi_get_sysid(i) :
+		ptes[i].part_table->sys_ind);
 }
 
 static struct systypes *
 get_sys_types(void) {
 	return (
 		sun_label ? sun_sys_types :
-		sgi_label ? sgi_sys_types : i386_sys_types);
+		sgi_label ? sgi_sys_types :
+		i386_sys_types);
 }
 
 char *partition_type(unsigned char type)
@@ -494,6 +506,13 @@ void list_types(struct systypes *sys)
 	putchar('\n');
 }
 
+static int
+is_cleared_partition(struct partition *p) {
+	return !(!p || p->boot_ind || p->head || p->sector || p->cyl ||
+		 p->sys_ind || p->end_head || p->end_sector || p->end_cyl ||
+		 get_start_sect(p) || get_nr_sects(p));
+}
+
 static void
 clear_partition(struct partition *p) {
 	if (!p)
@@ -511,8 +530,17 @@ clear_partition(struct partition *p) {
 }
 
 static void
-set_partition(int i, struct partition *p, uint start, uint stop,
-	      int sysid, uint offset) {
+set_partition(int i, int doext, uint start, uint stop, int sysid) {
+	struct partition *p;
+	uint offset;
+
+	if (doext) {
+		p = ptes[i].ext_pointer;
+		offset = extended_offset;
+	} else {
+		p = ptes[i].part_table;
+		offset = ptes[i].offset;
+	}
 	p->boot_ind = 0;
 	p->sys_ind = sysid;
 	set_start_sect(p, start - offset);
@@ -570,13 +598,13 @@ void update_units(void)
 static void
 warn_cylinders(void) {
 	if (dos_label && cylinders > 1024 && !nowarn)
-		fprintf(stderr, "\n\
-The number of cylinders for this disk is set to %d.\n\
-There is nothing wrong with that, but this is larger than 1024,\n\
-and could in certain setups cause problems with:\n\
-1) software that runs at boot time (e.g., old versions of LILO)\n\
-2) booting and partitioning software from other OSs\n\
-   (e.g., DOS FDISK, OS/2 FDISK)\n",
+		fprintf(stderr, _("\n"
+"The number of cylinders for this disk is set to %d.\n"
+"There is nothing wrong with that, but this is larger than 1024,\n"
+"and could in certain setups cause problems with:\n"
+"1) software that runs at boot time (e.g., old versions of LILO)\n"
+"2) booting and partitioning software from other OSs\n"
+"   (e.g., DOS FDISK, OS/2 FDISK)\n"),
 			cylinders);
 }
 
@@ -623,17 +651,18 @@ read_extended(int ext) {
 		for (i = 0; i < 4; i++, p++) if (get_nr_sects(p)) {
 			if (IS_EXTENDED (p->sys_ind)) {
 				if (pe->ext_pointer)
-					fprintf(stderr, _("Warning: extra link"
-							  " pointer in partition table "
-							  "%d\n"), partitions + 1);
+					fprintf(stderr,
+						_("Warning: extra link "
+						  "pointer in partition table"
+						  " %d\n"), partitions + 1);
 				else
 					pe->ext_pointer = p;
 			} else if (p->sys_ind) {
 				if (pe->part_table)
 					fprintf(stderr,
-						_("Warning: ignoring extra data "
-						  "in partition table %d\n"),
-						partitions + 1);
+						_("Warning: ignoring extra "
+						  "data in partition table"
+						  " %d\n"), partitions + 1);
 				else
 					pe->part_table = p;
 			}
@@ -662,7 +691,8 @@ read_extended(int ext) {
 	for (i = 4; i < partitions; i++) {
 		struct pte *pe = &ptes[i];
 
-		if (!get_nr_sects(pe->part_table)) {
+		if (!get_nr_sects(pe->part_table) &&
+		    (partitions > 5 || ptes[4].part_table->sys_ind)) {
 			printf("omitting empty partition (%d)\n", i+1);
 			delete_partition(i);
 			goto remove; 	/* numbering changed */
@@ -678,17 +708,18 @@ create_doslabel(void) {
 	_("Building a new DOS disklabel. Changes will remain in memory only,\n"
 	  "until you decide to write them. After that, of course, the previous\n"
 	  "content won't be recoverable.\n\n"));
-
 	sun_nolabel();  /* otherwise always recognised as sun */
 	sgi_nolabel();  /* otherwise always recognised as sgi */
+	aix_label = osf_label = 0;
+	partitions = 4;
 
+	for (i = 510-64; i < 510; i++)
+		MBRbuffer[i] = 0;
 	write_part_table_flag(MBRbuffer);
-	for (i = 0; i < 4; i++)
-	    if (ptes[i].part_table)
-		clear_partition(ptes[i].part_table);
+	extended_offset = 0;
 	set_all_unchanged();
 	set_changed(0);
-	get_boot(create_empty);
+	get_boot(create_empty_dos);
 }
 
 #include <sys/utsname.h>
@@ -727,52 +758,82 @@ get_sectorsize(int fd) {
 #endif
 }
 
-/*
- * Ask kernel about geometry. Invent something reasonable
- * in case the kernel has no opinion.
- */
+static void
+get_kernel_geometry(int fd) {
+#ifdef HDIO_GETGEO
+	struct hd_geometry geometry;
+
+	if (!ioctl(fd, HDIO_GETGEO, &geometry)) {
+		kern_heads = geometry.heads;
+		kern_sectors = geometry.sectors;
+		/* never use geometry.cylinders - it is truncated */
+	}
+#endif
+}
+
+static void
+get_partition_table_geometry(void) {
+	unsigned char *bufp = MBRbuffer;
+	struct partition *p;
+	int i, h, s, hh, ss;
+	int first = 1;
+	int bad = 0;
+
+	if (!(valid_part_table_flag(bufp)))
+		return;
+
+	hh = ss = 0;
+	for (i=0; i<4; i++) {
+		p = pt_offset(bufp, i);
+		if (p->sys_ind != 0) {
+			h = p->end_head + 1;
+			s = (p->end_sector & 077);
+			if (first) {
+				hh = h;
+				ss = s;
+				first = 0;
+			} else if (hh != h || ss != s)
+				bad = 1;
+		}
+	}
+
+	if (!first && !bad) {
+		pt_heads = hh;
+		pt_sectors = ss;
+	}
+}
+
 void
 get_geometry(int fd, struct geom *g) {
 	int sec_fac;
 	long longsectors;
-	struct hd_geometry geometry;
-	int res1, res2;
 
 	get_sectorsize(fd);
 	sec_fac = sector_size / 512;
-
 	guess_device_type(fd);
-
-	res1 = ioctl(fd, BLKGETSIZE, &longsectors);
-#ifdef HDIO_REQ
-	res2 = ioctl(fd, HDIO_REQ, &geometry);
-#else
-	res2 = ioctl(fd, HDIO_GETGEO, &geometry);
-#endif
-
-	/* never use geometry.cylinders - it is truncated */
 	heads = cylinders = sectors = 0;
-	sector_offset = 1;
-	if (res2 == 0) {
-		heads = geometry.heads;
-		sectors = geometry.sectors;
-		if (heads * sectors == 0)
-			res2 = -1;
-		else if (dos_compatible_flag)
-			sector_offset = sectors;
-	}
-	if (res1 == 0 && res2 == 0) { 	/* normal case */
-		cylinders = longsectors / (heads * sectors);
-		cylinders /= sec_fac;   /* do not round up */
-	} else if (res1 == 0) {		/* size but no geometry */
-		heads = cylinders = 1;
-		sectors = longsectors / sec_fac;
-	}
+	kern_heads = kern_sectors = 0;
+	pt_heads = pt_sectors = 0;
 
-	if (!sectors)
-		sectors = user_sectors;
-	if (!heads)
-		heads = user_heads;
+	get_kernel_geometry(fd);
+	get_partition_table_geometry();
+
+	heads = user_heads ? user_heads :
+		pt_heads ? pt_heads :
+		kern_heads ? kern_heads : 255;
+	sectors = user_sectors ? user_sectors :
+		pt_sectors ? pt_sectors :
+		kern_sectors ? kern_sectors : 63;
+
+	if (ioctl(fd, BLKGETSIZE, &longsectors))
+		longsectors = 0;
+
+	sector_offset = 1;
+	if (dos_compatible_flag)
+		sector_offset = sectors;
+
+	cylinders = longsectors / (heads * sectors);
+	cylinders /= sec_fac;
 	if (!cylinders)
 		cylinders = user_cylinders;
 
@@ -787,6 +848,7 @@ get_geometry(int fd, struct geom *g) {
  * Read MBR.  Returns:
  *   -1: no 0xaa55 flag present (possibly entire disk BSD)
  *    0: found or created label
+ *    1: I/O error
  */
 int
 get_boot(enum action what) {
@@ -794,23 +856,30 @@ get_boot(enum action what) {
 
 	partitions = 4;
 
-	if (what == create_empty)
-		goto got_table;		/* skip reading disk */
+	if (what == create_empty_dos)
+		goto got_dos_table;		/* skip reading disk */
+	if (what == create_empty_sun)
+		goto got_table;
 
 	if ((fd = open(disk_device, type_open)) < 0) {
-	    if ((fd = open(disk_device, O_RDONLY)) < 0)
+	    if ((fd = open(disk_device, O_RDONLY)) < 0) {
+		if (what == try_only)
+		    return 1;
 		fatal(unable_to_open);
-	    else
+	    } else
 		printf(_("You will not be able to write the partition table.\n"));
 	}
 
+	if (512 != read(fd, MBRbuffer, 512)) {
+		if (what == try_only)
+			return 1;
+		fatal(unable_to_read);
+	}
+
 	get_geometry(fd, NULL);
 
 	update_units();
 
-	if (sector_size != read(fd, MBRbuffer, sector_size))
-		fatal(unable_to_read);
-
 got_table:
 
 	if (check_sun_label())
@@ -825,6 +894,8 @@ got_table:
 	if (check_osf_label())
 		return 0;
 
+got_dos_table:
+
 	if (!valid_part_table_flag(MBRbuffer)) {
 		switch(what) {
 		case fdisk:
@@ -841,7 +912,8 @@ got_table:
 			return -1;
 		case try_only:
 		        return -1;
-		case create_empty:
+		case create_empty_dos:
+		case create_empty_sun:
 			break;
 		}
 
@@ -857,9 +929,9 @@ got_table:
 
 		pe->part_table = pt_offset(MBRbuffer, i);
 		pe->ext_pointer = NULL;
-		pe->changed = 0;
 		pe->offset = 0;
 		pe->sectorbuffer = MBRbuffer;
+		pe->changed = (what == create_empty_dos);
 	}
 
 	for (i = 0; i < 4; i++) {
@@ -925,6 +997,7 @@ char
 read_chars(char *mesg)
 {
         fputs(mesg, stdout);
+	fflush (stdout);	/* niles@scyld.com */
         if (!read_line()) {
 		*line_ptr = '\n';
 		line_ptr[1] = 0;
@@ -968,8 +1041,8 @@ read_int(uint low, uint dflt, uint high, uint base, char *mesg)
 	static char *ms = NULL;
 	static int mslen = 0;
 
-	if (!ms || strlen(mesg)+50 > mslen) {
-		mslen = strlen(mesg)+100;
+	if (!ms || strlen(mesg)+100 > mslen) {
+		mslen = strlen(mesg)+200;
 		if (!(ms = realloc(ms,mslen)))
 			fatal(out_of_memory);
 	}
@@ -978,9 +1051,11 @@ read_int(uint low, uint dflt, uint high, uint base, char *mesg)
 		default_ok = 0;
 
 	if (default_ok)
-		sprintf(ms, _("%s (%d-%d, default %d): "), mesg, low, high, dflt);
+		snprintf(ms, mslen, _("%s (%d-%d, default %d): "),
+			 mesg, low, high, dflt);
 	else
-		sprintf(ms, "%s (%d-%d): ", mesg, low, high);
+		snprintf(ms, mslen, "%s (%d-%d): ",
+			 mesg, low, high);
 
 	while (1) {
 		int use_default = default_ok;
@@ -1049,13 +1124,13 @@ get_partition(int warn, int max) {
 	i = read_int(1, 0, max, 0, _("Partition number")) - 1;
 	pe = &ptes[i];
 
-	if (warn && (
-	     (!sun_label && !sgi_label && !pe->part_table->sys_ind)
+	if (warn
+	    && ((!sun_label && !sgi_label && !pe->part_table->sys_ind)
 	     || (sun_label &&
 		(!sunlabel->partitions[i].num_sectors ||
 		 !sunlabel->infos[i].id))
-	     || (sgi_label && (!sgi_get_num_sectors(i)))
-	)) fprintf(stderr, _("Warning: partition %d has empty type\n"), i+1);
+		|| (sgi_label && (!sgi_get_num_sectors(i))))
+	) fprintf(stderr, _("Warning: partition %d has empty type\n"), i+1);
 	return i;
 }
 
@@ -1112,7 +1187,7 @@ delete_partition(int i) {
  */
 
 	if (warn_geometry())
-		return;
+		return;		/* C/H/S not set */
 	pe->changed = 1;
 
 	if (sun_label) {
@@ -1136,25 +1211,22 @@ delete_partition(int i) {
 	}
 
 	if (!q->sys_ind && i > 4) {
+		/* the last one in the chain - just delete */
 		--partitions;
 		--i;
 		clear_partition(ptes[i].ext_pointer);
 		ptes[i].changed = 1;
 	} else {
+		/* not the last one - further ones will be moved down */
 		if (i > 4) {
+			/* delete this link in the chain */
 			p = ptes[i-1].ext_pointer;
-			p->boot_ind = 0;
-			p->head = q->head;
-			p->sector = q->sector;
-			p->cyl = q->cyl;
-			p->sys_ind = EXTENDED;
-			p->end_head = q->end_head;
-			p->end_sector = q->end_sector;
-			p->end_cyl = q->end_cyl;
+			*p = *q;
 			set_start_sect(p, get_start_sect(q));
 			set_nr_sects(p, get_nr_sects(q));
 			ptes[i-1].changed = 1;
 		} else if (partitions > 5) {    /* 5 will be moved to 4 */
+			/* the first logical in a longer chain */
 			struct pte *pe = &ptes[5];
 
 			if(pe->part_table) /* prevent SEGFAULT */
@@ -1172,6 +1244,7 @@ delete_partition(int i) {
 				i++;
 			}
 		} else
+			/* the only logical: clear only */
 			clear_partition(ptes[i].part_table);
 	}
 }
@@ -1184,12 +1257,12 @@ change_sysid(void) {
 
 	origsys = sys = get_sysid(i);
 
-	if (!sys && !sgi_label)
+	if (!sys && !sgi_label && !sun_label)
                 printf(_("Partition %d does not exist yet!\n"), i + 1);
         else while (1) {
 		sys = read_hex (get_sys_types());
 
-		if (!sys && !sgi_label) {
+		if (!sys && !sgi_label && !sun_label) {
 			printf(_("Type 0 means free space to many systems\n"
 			       "(but not to Linux). Having partitions of\n"
 			       "type 0 is probably unwise. You can delete\n"
@@ -1220,7 +1293,6 @@ change_sysid(void) {
 				       "as IRIX expects it.\n\n"));
                         if (sys == origsys)
                             break;
-
 			if (sun_label) {
 				sun_change_sysid(i, sys);
 			} else
@@ -1233,6 +1305,9 @@ change_sysid(void) {
                                 (temp = partition_type(sys)) ? temp :
                                 _("Unknown"));
                         ptes[i].changed = 1;
+			if (is_dos_partition(origsys) ||
+			    is_dos_partition(sys))
+				dos_changed = 1;
                         break;
                 }
         }
@@ -1354,9 +1429,72 @@ wrong_p_order(int *prev) {
 	return 0;
 }
 
+/*
+ * Fix the chain of logicals.
+ * extended_offset is unchanged, the set of sectors used is unchanged
+ * The chain is sorted so that sectors increase, and so that
+ * starting sectors increase.
+ *
+ * After this it may still be that cfdisk doesnt like the table.
+ * (This is because cfdisk considers expanded parts, from link to
+ * end of partition, and these may still overlap.)
+ * Now
+ *   sfdisk /dev/hda > ohda; sfdisk /dev/hda < ohda
+ * may help.
+ */
+static void
+fix_chain_of_logicals(void) {
+	int j, oj, ojj, sj, sjj;
+	struct partition *pj,*pjj,tmp;
+
+	/* Stage 1: sort sectors but leave sector of part 4 */
+	/* (Its sector is the global extended_offset.) */
+ stage1:
+	for(j = 5; j < partitions-1; j++) {
+		oj = ptes[j].offset;
+		ojj = ptes[j+1].offset;
+		if (oj > ojj) {
+			ptes[j].offset = ojj;
+			ptes[j+1].offset = oj;
+			pj = ptes[j].part_table;
+			set_start_sect(pj, get_start_sect(pj)+oj-ojj);
+			pjj = ptes[j+1].part_table;
+			set_start_sect(pjj, get_start_sect(pjj)+ojj-oj);
+			set_start_sect(ptes[j-1].ext_pointer,
+				       ojj-extended_offset);
+			set_start_sect(ptes[j].ext_pointer,
+				       oj-extended_offset);
+			goto stage1;
+		}
+	}
+
+	/* Stage 2: sort starting sectors */
+ stage2:
+	for(j = 4; j < partitions-1; j++) {
+		pj = ptes[j].part_table;
+		pjj = ptes[j+1].part_table;
+		sj = get_start_sect(pj);
+		sjj = get_start_sect(pjj);
+		oj = ptes[j].offset;
+		ojj = ptes[j+1].offset;
+		if (oj+sj > ojj+sjj) {
+			tmp = *pj;
+			*pj = *pjj;
+			*pjj = tmp;
+			set_start_sect(pj, ojj+sjj-oj);
+			set_start_sect(pjj, oj+sj-ojj);
+			goto stage2;
+		}
+	}
+
+	/* Probably something was changed */
+	for(j = 4; j < partitions; j++)
+		ptes[j].changed = 1;
+}
+
 static void
 fix_partition_table_order(void) {
-	struct pte *pei, *pek, ptebuf;
+	struct pte *pei, *pek;
 	int i,k;
 
 	if(!wrong_p_order(NULL)) {
@@ -1364,34 +1502,32 @@ fix_partition_table_order(void) {
 		return;
 	}
 
-	while ((i = wrong_p_order(&k)) != 0) {
+	while ((i = wrong_p_order(&k)) != 0 && i < 4) {
 		/* partition i should have come earlier, move it */
+		/* We have to move data in the MBR */
+		struct partition *pi, *pk, *pe, pbuf;
 		pei = &ptes[i];
 		pek = &ptes[k];
 
-		if (i < 4) {
-			/* We have to move data in the MBR */
-			struct partition *pi, *pk, *pe, pbuf;
+		pe = pei->ext_pointer;
+		pei->ext_pointer = pek->ext_pointer;
+		pek->ext_pointer = pe;
 
-			pe = pei->ext_pointer;
-			pei->ext_pointer = pek->ext_pointer;
-			pek->ext_pointer = pe;
+		pi = pei->part_table;
+		pk = pek->part_table;
 
-			pi = pei->part_table;
-			pk = pek->part_table;
+		memmove(&pbuf, pi, sizeof(struct partition));
+		memmove(pi, pk, sizeof(struct partition));
+		memmove(pk, &pbuf, sizeof(struct partition));
 
-			memmove(&pbuf, pi, sizeof(struct partition));
-			memmove(pi, pk, sizeof(struct partition));
-			memmove(pk, &pbuf, sizeof(struct partition));
-		} else {
-			/* Only change is change in numbering */
-			ptebuf = *pei;
-			*pei = *pek;
-			*pek = ptebuf;
-		}
 		pei->changed = pek->changed = 1;
-
 	}
+
+	if (i)
+		fix_chain_of_logicals();
+
+	printf("Done.\n");
+
 }
 
 static void
@@ -1433,7 +1569,7 @@ list_table(int xtra) {
 		struct pte *pe = &ptes[i];
 
 		p = pe->part_table;
-		if (p->sys_ind) {
+		if (p && !is_cleared_partition(p)) {
 			unsigned int psects = get_nr_sects(p);
 			unsigned int pblocks = psects;
 			unsigned int podd = 0;
@@ -1459,7 +1595,7 @@ list_table(int xtra) {
 			check_consistency(p, i);
 		}
 	}
-	
+
 	/* Is partition table in disk order? It need not be, but... */
 	/* partition table entries are not checked for correct order if this
 	   is a sgi, sun or aix labeled disk... */
@@ -1542,7 +1678,7 @@ verify(void) {
 	int i, j;
 	uint total = 1;
 	uint first[partitions], last[partitions];
-	struct partition *p = part_table[0];
+	struct partition *p;
 
 	if (warn_geometry())
 		return;
@@ -1621,7 +1757,7 @@ add_partition(int n, int sys) {
 
 	if (p && p->sys_ind) {
 		printf(_("Partition %d is already defined.  Delete "
-			"it before re-adding it.\n"), n + 1);
+			 "it before re-adding it.\n"), n + 1);
 		return;
 	}
 	fill_bounds(first, last);
@@ -1641,7 +1777,7 @@ add_partition(int n, int sys) {
 		for (i = 0; i < partitions; i++)
 			first[i] = (cround(first[i]) - 1) * units_per_sector;
 
-	sprintf(mesg, _("First %s"), str_units(SINGULAR));
+	snprintf(mesg, sizeof(mesg), _("First %s"), str_units(SINGULAR));
 	do {
 		temp = start;
 		for (i = 0; i < partitions; i++) {
@@ -1677,9 +1813,9 @@ add_partition(int n, int sys) {
 
 		pe->offset = start - sector_offset;
 		if (pe->offset == extended_offset) { /* must be corrected */
-		     pe->offset++;
-		     if (sector_offset == 1)
-			  start++;
+			pe->offset++;
+			if (sector_offset == 1)
+				start++;
 		}
 	}
 
@@ -1700,8 +1836,9 @@ add_partition(int n, int sys) {
 	if (cround(start) == cround(limit)) {
 		stop = limit;
 	} else {
-		sprintf(mesg, _("Last %s or +size or +sizeM or +sizeK"),
-			str_units(SINGULAR));
+		snprintf(mesg, sizeof(mesg),
+			 _("Last %s or +size or +sizeM or +sizeK"),
+			 str_units(SINGULAR));
 		stop = read_int(cround(start), cround(limit), cround(limit),
 				cround(start), mesg);
 		if (display_in_cyl_units) {
@@ -1711,7 +1848,9 @@ add_partition(int n, int sys) {
 		}
 	}
 
-	set_partition(n, p, start, stop, sys, ptes[n].offset);
+	set_partition(n, 0, start, stop, sys);
+	if (n > 4)
+		set_partition(n - 1, 1, ptes[n].offset, stop, EXTENDED);
 
 	if (IS_EXTENDED (sys)) {
 		struct pte *pe4 = &ptes[4];
@@ -1726,16 +1865,6 @@ add_partition(int n, int sys) {
 		pe4->ext_pointer = pe4->part_table + 1;
 		pe4->changed = 1;
 		partitions = 5;
-	} else {
-		if (n > 4)
-			set_partition(n - 1, ptes[n-1].ext_pointer,
-				ptes[n].offset, stop, EXTENDED,
-				extended_offset);
-#if 0
-		if ((limit = get_nr_sects(p)) & 1)
-			printf(_("Warning: partition %d has an odd "
-				"number of sectors.\n"), n + 1);
-#endif
 	}
 }
 
@@ -1772,6 +1901,15 @@ new_partition(void) {
 		return;
 	}
 
+	if (aix_label) {
+		printf(_("\tSorry - this fdisk cannot handle AIX disk labels."
+			 "\n\tIf you want to add DOS-type partitions, create"
+			 "\n\ta new empty DOS partition table first. (Use o.)"
+			 "\n\tWARNING: "
+			 "This will destroy the present disk contents.\n"));
+		return;
+	}
+
 	if (partitions >= MAXIMUM_PARTS) {
 		printf(_("The maximum number of partitions has been created\n"));
 		return;
@@ -1784,16 +1922,17 @@ new_partition(void) {
 			add_logical();
 		else
 			printf(_("You must delete some partition and add "
-				"an extended partition first\n"));
+				 "an extended partition first\n"));
 	} else {
 		char c, line[LINE_LENGTH];
-		sprintf(line, _("Command action\n   %s\n   p   primary "
-			"partition (1-4)\n"), extended_offset ?
-			_("l   logical (5 or over)") : _("e   extended"));
-		while (1)
+		snprintf(line, sizeof(line),
+			 _("Command action\n   %s\n   p   primary "
+			   "partition (1-4)\n"), extended_offset ?
+			 _("l   logical (5 or over)") : _("e   extended"));
+		while (1) {
 			if ((c = tolower(read_char(line))) == 'p') {
 				add_partition(get_partition(0, 4),
-					LINUX_NATIVE);
+					      LINUX_NATIVE);
 				return;
 			}
 			else if (c == 'l' && extended_offset) {
@@ -1802,13 +1941,13 @@ new_partition(void) {
 			}
 			else if (c == 'e' && !extended_offset) {
 				add_partition(get_partition(0, 4),
-					EXTENDED);
+					      EXTENDED);
 				return;
 			}
 			else
 				printf(_("Invalid partition number "
-				       "for type `%c'\n"), c);
-		
+					 "for type `%c'\n"), c);
+		}
 	}
 }
 
@@ -1828,7 +1967,8 @@ write_table(void) {
 				write_sector(fd, pe->offset, pe->sectorbuffer);
 			}
 		}
-	} else if (sgi_label) {
+	}
+	else if (sgi_label) {
 		/* no test on change? the printf below might be mistaken */
 		sgi_write_table();
 	} else if (sun_label) {
@@ -1865,12 +2005,16 @@ reread_partition_table(int leave) {
                         error = errno;
         }
 
-	if (i < 0)
-		printf(_("Re-read table failed with error %d: %s.\nReboot your "
-			"system to ensure the partition table is updated.\n"),
+	if (i) {
+		printf(_("\nWARNING: Re-reading the partition table "
+			 "failed with error %d: %s.\n"
+			 "The kernel still uses the old table.\n"
+			 "The new table will be used "
+			 "at the next reboot.\n"),
 			error, strerror(error));
+	}
 
-	if (!sun_label && !sgi_label)
+	if (dos_changed)
 	    printf(
 		_("\nWARNING: If you have created or modified any DOS 6.x\n"
 		"partitions, please see the fdisk manual page for additional\n"
@@ -1959,7 +2103,7 @@ xselect(void) {
 			break;
 		case 'c':
 			user_cylinders = cylinders =
-				read_int(1, cylinders, 131071, 0,
+				read_int(1, cylinders, 1048576, 0,
 					 _("Number of cylinders"));
 			if (sun_label)
 				sun_set_ncyl(cylinders);
@@ -1974,7 +2118,8 @@ xselect(void) {
 				sgi_set_xcyl();
 			else if (sun_label)
 				sun_set_xcyl();
-			else if (dos_label)
+			else
+			if (dos_label)
 				x_list_table(1);
 			break;
 		case 'f':
@@ -2037,64 +2182,74 @@ xselect(void) {
 }
 
 static int
-is_ide_cdrom(char *device) {
+is_ide_cdrom_or_tape(char *device) {
+	FILE *procf;
+	char buf[100];
+	struct stat statbuf;
+	int is_ide = 0;
+
 	/* No device was given explicitly, and we are trying some
-       likely things.  But opening /dev/hdc may produce errors like
+	   likely things.  But opening /dev/hdc may produce errors like
            "hdc: tray open or drive not ready"
-       if it happens to be a CD-ROM drive. It even happens that
-       the process hangs on the attempt to read a music CD.
-       So try to be careful. This only works since 2.1.73. */
+	   if it happens to be a CD-ROM drive. It even happens that
+	   the process hangs on the attempt to read a music CD.
+	   So try to be careful. This only works since 2.1.73. */
 
-    FILE *procf;
-    char buf[100];
-    struct stat statbuf;
+	if (strncmp("/dev/hd", device, 7))
+		return 0;
 
-    if (strncmp("/dev/hd", device, 7))
-	return 0;
-    sprintf(buf, "/proc/ide/%s/media", device+5);
-    procf = fopen(buf, "r");
-    if (procf != NULL && fgets(buf, sizeof(buf), procf))
-        return  !strncmp(buf, "cdrom", 5);
+	snprintf(buf, sizeof(buf), "/proc/ide/%s/media", device+5);
+	procf = fopen(buf, "r");
+	if (procf != NULL && fgets(buf, sizeof(buf), procf))
+		is_ide = (!strncmp(buf, "cdrom", 5) ||
+			  !strncmp(buf, "tape", 4));
+	else
+		/* Now when this proc file does not exist, skip the
+		   device when it is read-only. */
+		if (stat(device, &statbuf) == 0)
+			is_ide = ((statbuf.st_mode & 0222) == 0);
 
-    /* Now when this proc file does not exist, skip the
-       device when it is read-only. */
-    if (stat(device, &statbuf) == 0)
-        return (statbuf.st_mode & 0222) == 0;
-
-    return 0;
+	if (procf)
+		fclose(procf);
+	return is_ide;
 }
 
 static void
 try(char *device, int user_specified) {
+	int gb;
+
 	disk_device = device;
-	if (!setjmp(listingbuf)) {
-		if (!user_specified)
-			if (is_ide_cdrom(device))
+	if (setjmp(listingbuf))
+		return;
+	if (!user_specified)
+		if (is_ide_cdrom_or_tape(device))
+			return;
+	if ((fd = open(disk_device, type_open)) >= 0) {
+		gb = get_boot(try_only);
+		if (gb > 0) {	/* I/O error */
+			close(fd);
+		} else if (gb < 0) { /* no DOS signature */
+			list_disk_geometry();
+			if (aix_label)
 				return;
-		if ((fd = open(disk_device, type_open)) >= 0) {
-			if (get_boot(try_only) < 0) {
-				list_disk_geometry();
-				if (aix_label)
-					return;
-				if (btrydev(device) < 0)
-					fprintf(stderr,
-						_("Disk %s doesn't contain a valid "
-						"partition table\n"), device);
-				close(fd);
-			} else {
-				close(fd);
-				list_table(0);
-				if (!sun_label && partitions > 4)
-					delete_partition(ext_index);
-			}
+			if (btrydev(device) < 0)
+				fprintf(stderr,
+					_("Disk %s doesn't contain a valid "
+					  "partition table\n"), device);
+			close(fd);
 		} else {
-				/* Ignore other errors, since we try IDE
-				   and SCSI hard disks which may not be
-				   installed on the system. */
-			if(errno == EACCES) {
-			    fprintf(stderr, _("Cannot open %s\n"), device);
-			    return;
-			}
+			close(fd);
+			list_table(0);
+			if (!sun_label && partitions > 4)
+				delete_partition(ext_index);
+		}
+	} else {
+		/* Ignore other errors, since we try IDE
+		   and SCSI hard disks which may not be
+		   installed on the system. */
+		if(errno == EACCES) {
+			fprintf(stderr, _("Cannot open %s\n"), device);
+			return;
 		}
 	}
 }
@@ -2114,15 +2269,16 @@ tryprocpt(void) {
 	}
 
 	while (fgets(line, sizeof(line), procpt)) {
-		if (sscanf (line, " %d %d %d %[^\n]\n",
+		if (sscanf (line, " %d %d %d %[^\n ]",
 			    &ma, &mi, &sz, ptname) != 4)
 			continue;
 		for(s = ptname; *s; s++);
 		if (isdigit(s[-1]))
 			continue;
-		sprintf(devname, "/dev/%s", ptname);
+		snprintf(devname, sizeof(devname), "/dev/%s", ptname);
 		try(devname, 0);
 	}
+	fclose(procpt);
 }
 
 static void
@@ -2149,7 +2305,7 @@ main(int argc, char **argv) {
 	 *  fdisk -s [partition] ...
 	 *  fdisk [-b sectorsize] [-u] device
 	 */
-	while ((c = getopt(argc, argv, "b:lsuvV")) != EOF) {
+	while ((c = getopt(argc, argv, "b:lsuvV")) != -1) {
 		switch (c) {
 		case 'b':
 			/* ugly: this sector size is really per device,
@@ -2239,8 +2395,6 @@ main(int argc, char **argv) {
 		fatal(usage2);
 
 	get_boot(fdisk);
-	if (aix_label)
-		exit(0);
 
 #ifdef __alpha__
 	/* On alpha, if we detect a disklabel, go directly to
@@ -2254,7 +2408,6 @@ main(int argc, char **argv) {
 		bselect();
 	}
 #endif
-
 	while (1) {
 		putchar('\n');
 		c = tolower(read_char(_("Command (m for help): ")));
