@@ -26,7 +26,6 @@
 #undef u_char
 #endif
 #include <linux/major.h>	/* FLOPPY_MAJOR */
-#include <linux/hdreg.h>	/* HDIO_GETGEO */
 
 #include "common.h"
 #include "fdisk.h"
@@ -51,6 +50,10 @@ struct systypes sun_sys_types[] = {
 	{8, N_("SunOS home")},
 	{LINUX_SWAP, N_("Linux swap")},
 	{LINUX_NATIVE, N_("Linux native")},
+	{0x8e, N_("Linux LVM")},
+	{0xfd, N_("Linux raid autodetect")},/* New (2.2.x) raid partition
+					       with autodetect using
+					       persistent superblock */
 	{ 0, NULL }
 };
 
@@ -66,6 +69,12 @@ get_num_sectors(struct sun_partition p) {
 	return SSWAP32(p.num_sectors);
 }
 
+#ifndef IDE0_MAJOR
+#define IDE0_MAJOR 3
+#endif
+#ifndef IDE1_MAJOR
+#define IDE1_MAJOR 22
+#endif
 void guess_device_type(int fd) {
 	struct stat bootstat;
 
@@ -87,25 +96,25 @@ void guess_device_type(int fd) {
 	}
 }
 
-void set_sun_partition(int i, uint start, uint stop, int sysid)
-{
+static void
+set_sun_partition(int i, uint start, uint stop, int sysid) {
 	sunlabel->infos[i].id = sysid;
 	sunlabel->partitions[i].start_cylinder =
 		SSWAP32(start / (heads * sectors));
 	sunlabel->partitions[i].num_sectors =
 		SSWAP32(stop - start);
-	changed[i] = 1;
+	set_changed(i);
 }
 
-void sun_nolabel(void)
-{
+void
+sun_nolabel(void) {
 	sun_label = 0;
 	sunlabel->magic = 0;
 	partitions = 4;
 }
 
-int check_sun_label(void)
-{
+int
+check_sun_label(void) {
 	unsigned short *ush;
 	int csum;
 	
@@ -148,6 +157,10 @@ struct sun_predefined_drives {
 {"Quantum","ProDrive 80S",1,832,2,834,6,34,3662},
 {"Quantum","ProDrive 105S",1,974,2,1019,6,35,3662},
 {"CDC","Wren IV 94171-344",3,1545,2,1549,9,46,3600},
+{"IBM","DPES-31080",0,4901,2,4903,4,108,5400},
+{"IBM","DORS-32160",0,1015,2,1017,67,62,5400},
+{"IBM","DNES-318350",0,11199,2,11474,10,320,7200},
+{"SEAGATE","ST34371",0,3880,2,3882,16,135,7228},
 {"","SUN0104",1,974,2,1019,6,35,3662},
 {"","SUN0207",4,1254,2,1272,9,36,3600},
 {"","SUN0327",3,1545,2,1549,9,46,3600},
@@ -162,8 +175,8 @@ struct sun_predefined_drives {
 {"IOMEGA","Jaz",0,1019,2,1021,64,32,5394},
 };
 
-struct sun_predefined_drives *
-sun_autoconfigure_scsi() {
+static struct sun_predefined_drives *
+sun_autoconfigure_scsi(void) {
     struct sun_predefined_drives *p = NULL;
 
 #ifdef SCSI_IOCTL_GET_IDLUN
@@ -193,16 +206,18 @@ sun_autoconfigure_scsi() {
         pfd = fopen("/proc/scsi/scsi","r");
         if (pfd) {
             while (fgets(buffer2,2048,pfd)) {
-                            if (!strcmp(buffer, buffer2)) {
+		if (!strcmp(buffer, buffer2)) {
                     if (fgets(buffer2,2048,pfd)) {
                         q = strstr(buffer2,"Vendor: ");
                         if (q) {
                             q += 8;
                             vendor = q;
-                            q = strstr(q," Model: ");
+			    q = strstr(q," ");
+			    *q++ = 0;	/* truncate vendor name */
+                            q = strstr(q,"Model: ");
                             if (q) {
                                 *q = 0;
-                                q += 8;
+                                q += 7;
                                 model = q;
                                 q = strstr(q," Rev: ");
                                 if (q) {
@@ -247,7 +262,7 @@ void create_sunlabel(void)
 #else
 	other_endian = 0;
 #endif
-	memset(buffer, 0, SECTOR_SIZE);
+	memset(MBRbuffer, 0, sizeof(MBRbuffer));
 	sunlabel->magic = SSWAP16(SUN_LABEL_MAGIC);
 	if (!floppy) {
 	    puts(_("Drive type\n"
@@ -328,14 +343,20 @@ void create_sunlabel(void)
 	    sunlabel->ntrks = SSWAP16(p->ntrks);
 	    sunlabel->nsect = SSWAP16(p->nsect);
 	    sunlabel->rspeed = SSWAP16(p->rspeed);
+	    sunlabel->ilfact = SSWAP16(1);
 	    cylinders = p->ncyl;
 	    heads = p->ntrks;
 	    sectors = p->nsect;
 	    puts(_("You may change all the disk params from the x menu"));
 	}
-	sprintf(buffer, "%s%s%s cyl %d alt %d hd %d sec %d", 
-			p ? p->vendor : "", (p && *p->vendor) ? " " : "", p ? p->model : (floppy ? _("3,5\" floppy") : _("Linux custom")),
-			cylinders, SSWAP16(sunlabel->nacyl), heads, sectors);
+
+	snprintf(sunlabel->info, sizeof(sunlabel->info),
+		 "%s%s%s cyl %d alt %d hd %d sec %d", 
+		 p ? p->vendor : "", (p && *p->vendor) ? " " : "",
+		 p ? p->model
+		   : (floppy ? _("3,5\" floppy") : _("Linux custom")),
+		cylinders, SSWAP16(sunlabel->nacyl), heads, sectors);
+
 	sunlabel->ntrks = SSWAP16(heads);
 	sunlabel->nsect = SSWAP16(sectors);
 	sunlabel->ncyl = SSWAP16(cylinders);
@@ -358,26 +379,28 @@ void create_sunlabel(void)
 			csum ^= *ush++;
 		sunlabel->csum = csum;
 	}
-	for (i = 1; i < MAXIMUM_PARTS; i++)
-	    changed[i] = 0;
-	changed[0] = 1;
-	get_boot(create_empty);
+
+	set_all_unchanged();
+	set_changed(0);
+	get_boot(create_empty_sun);
 }
 
-void toggle_sunflags(int i, unsigned char mask)
-{
+void
+toggle_sunflags(int i, unsigned char mask) {
 	if (sunlabel->infos[i].flags & mask)
 		sunlabel->infos[i].flags &= ~mask;
 	else sunlabel->infos[i].flags |= mask;
-	changed[i] = 1;
+	set_changed(i);
 }
 
-void fetch_sun(uint *starts, uint *lens, uint *start, uint *stop)
-{
+static void
+fetch_sun(uint *starts, uint *lens, uint *start, uint *stop) {
 	int i, continuous = 1;
 	*start = 0; *stop = cylinders * heads * sectors;
 	for (i = 0; i < partitions; i++) {
-		if (sunlabel->partitions[i].num_sectors && sunlabel->infos[i].id && sunlabel->infos[i].id != WHOLE_DISK) {
+		if (sunlabel->partitions[i].num_sectors
+		    && sunlabel->infos[i].id
+		    && sunlabel->infos[i].id != WHOLE_DISK) {
 			starts[i] = SSWAP32(sunlabel->partitions[i].start_cylinder) * heads * sectors;
 			lens[i] = SSWAP32(sunlabel->partitions[i].num_sectors);
 			if (continuous) {
@@ -386,7 +409,9 @@ void fetch_sun(uint *starts, uint *lens, uint *start, uint *stop)
 				else if (starts[i] + lens[i] >= *stop)
 					*stop = starts[i];
 				else
-					continuous = 0; /* There will be probably more gaps than one, so lets check afterwards */
+					continuous = 0;
+				        /* There will be probably more gaps
+					  than one, so lets check afterwards */
 			}
 		} else {
 			starts[i] = 0;
@@ -396,16 +421,17 @@ void fetch_sun(uint *starts, uint *lens, uint *start, uint *stop)
 }
 
 static uint *verify_sun_starts;
-int verify_sun_cmp(int *a, int *b)
-{
+
+static int
+verify_sun_cmp(int *a, int *b) {
     if (*a == -1) return 1;
     if (*b == -1) return -1;
     if (verify_sun_starts[*a] > verify_sun_starts[*b]) return 1;
     return -1;
 }
 
-void verify_sun(void)
-{
+void
+verify_sun(void) {
     uint starts[8], lens[8], start, stop;
     int i,j,k,starto,endo;
     int array[8];
@@ -466,13 +492,13 @@ void verify_sun(void)
         printf(_("Unused gap - sectors %d-%d\n"),start,stop);
 }
 
-void add_sun_partition(int n, int sys)
-{
+void
+add_sun_partition(int n, int sys) {
 	uint start, stop, stop2;
 	uint starts[8], lens[8];
 	int whole_disk = 0;
 		
-	char mesg[48];
+	char mesg[256];
 	int i, first, last;
 
 	if (sunlabel->partitions[n].num_sectors && sunlabel->infos[n].id) {
@@ -491,7 +517,7 @@ void add_sun_partition(int n, int sys)
 			return;
 		}
 	}
-	sprintf(mesg, _("First %s"), str_units(SINGULAR));
+	snprintf(mesg, sizeof(mesg), _("First %s"), str_units(SINGULAR));
 	for (;;) {
 		if (whole_disk)
 			first = read_int(0, 0, 0, 0, mesg);
@@ -507,6 +533,22 @@ void add_sun_partition(int n, int sys)
 			printf ("\
 It is highly recommended that the third partition covers the whole disk\n\
 and is of type `Whole disk'\n");
+		/* ewt asks to add: "don't start a partition at cyl 0"
+		   However, edmundo@rano.demon.co.uk writes:
+		   "In addition to having a Sun partition table, to be able to
+		   boot from the disc, the first partition, /dev/sdX1, must
+		   start at cylinder 0. This means that /dev/sdX1 contains
+		   the partition table and the boot block, as these are the
+		   first two sectors of the disc. Therefore you must be
+		   careful what you use /dev/sdX1 for. In particular, you must
+		   not use a partition starting at cylinder 0 for Linux swap,
+		   as that would overwrite the partition table and the boot
+		   block. You may, however, use such a partition for a UFS
+		   or EXT2 file system, as these file systems leave the first
+		   1024 bytes undisturbed. */
+		/* On the other hand, one should not use partitions
+		   starting at block 0 in an md, or the label will
+		   be trashed. */
 		for (i = 0; i < partitions; i++)
 			if (lens[i] && starts[i] <= first
 			            && starts[i] + lens[i] > first)
@@ -526,7 +568,9 @@ and is of type `Whole disk'\n");
 		if (starts[i] > first && starts[i] < stop)
 			stop = starts[i];
 	}
-	sprintf(mesg, _("Last %s or +size or +sizeM or +sizeK"), str_units(SINGULAR));
+	snprintf(mesg, sizeof(mesg),
+		 _("Last %s or +size or +sizeM or +sizeK"),
+		 str_units(SINGULAR));
 	if (whole_disk)
 		last = read_int(scround(stop2), scround(stop2), scround(stop2),
 				0, mesg);
@@ -582,7 +626,7 @@ sun_change_sysid(int i, int sys) {
 	      "there may destroy your partition table and bootblock.\n"
 	      "Type YES if you're very sure you would like that partition\n"
 	      "tagged with 82 (Linux swap): "));
-	    if (strcmp (line_ptr, "YES\n"))
+	    if (strcmp (line_ptr, _("YES\n")))
 		    return;
 	}
 	switch (sys) {
@@ -701,4 +745,3 @@ sun_write_table(void) {
 	if (write(fd, sunlabel, SECTOR_SIZE) != SECTOR_SIZE)
 		fatal(unable_to_write);
 }
-
