@@ -47,7 +47,6 @@
 #include <sys/stat.h>
 #include <sys/utsname.h>
 #include <linux/unistd.h>	/* _syscall */
-#include <linux/hdreg.h>	/* HDIO_GETGEO */
 #include "nls.h"
 #include "common.h"
 
@@ -131,7 +130,7 @@ fatal(char *s, ...) {
  *
  * Note: we use 512-byte sectors here, irrespective of the hardware ss.
  */
-#if !defined (__alpha__) && !defined (__ia64__)
+#if !defined (__alpha__) && !defined (__ia64__) && !defined (__x86_64__)
 static
 _syscall5(int,  _llseek,  uint,  fd, ulong, hi, ulong, lo,
        loff_t *, res, uint, wh);
@@ -143,7 +142,7 @@ sseek(char *dev, unsigned int fd, unsigned long s) {
     in = ((loff_t) s << 9);
     out = 1;
 
-#if !defined (__alpha__) && !defined (__ia64__)
+#if !defined (__alpha__) && !defined (__ia64__) && !defined (__x86_64__)
     if (_llseek (fd, in>>32, in & 0xffffffff, &out, SEEK_SET) != 0) {
 #else
     if ((out = lseek(fd, in, SEEK_SET)) != in) {
@@ -220,12 +219,12 @@ get_sector(char *dev, int fd, unsigned long sno) {
 
 static int
 msdos_signature (struct sector *s) {
-    if (*(unsigned short *) (s->data + 0x1fe) != 0xaa55) {
-	error(_("ERROR: sector %lu does not have an msdos signature\n"),
-	       s->sectornumber);
-	return 0;
-    }
-    return 1;
+    unsigned char *data = s->data;
+    if (data[510] == 0x55 && data[511] == 0xaa)
+	    return 1;
+    error(_("ERROR: sector %lu does not have an msdos signature\n"),
+	  s->sectornumber);
+    return 0;
 }
 
 static int
@@ -432,25 +431,25 @@ get_cylindersize(char *dev, int fd, int silent) {
 
     if (R.start && !force) {
 	warn(
-	    _("Warning: start=%d - this looks like a partition rather than\n"
+	    _("Warning: start=%lu - this looks like a partition rather than\n"
 	      "the entire disk. Using fdisk on it is probably meaningless.\n"
 	      "[Use the --force option if you really want this]\n"), R.start);
 	exit(1);
     }
 
     if (R.heads && B.heads != R.heads)
-	warn(_("Warning: HDIO_GETGEO says that there are %d heads\n"),
+	warn(_("Warning: HDIO_GETGEO says that there are %lu heads\n"),
 	     R.heads);
     if (R.sectors && B.sectors != R.sectors)
-	warn(_("Warning: HDIO_GETGEO says that there are %d sectors\n"),
+	warn(_("Warning: HDIO_GETGEO says that there are %lu sectors\n"),
 	     R.sectors);
     if (R.cylinders && B.cylinders != R.cylinders
 	    && B.cylinders < 65536 && R.cylinders < 65536)
-	warn(_("Warning: BLKGETSIZE/HDIO_GETGEO says that there are %d cylinders\n"),
+	warn(_("Warning: BLKGETSIZE/HDIO_GETGEO says that there are %lu cylinders\n"),
 	     R.cylinders);
 
     if (B.sectors > 63)
-      warn(_("Warning: unlikely number of sectors (%d) - usually at most 63\n"
+      warn(_("Warning: unlikely number of sectors (%lu) - usually at most 63\n"
 	   "This will give problems with all software that uses C/H/S addressing.\n"),
 	   B.sectors);
     if (!silent)
@@ -537,17 +536,17 @@ chs_ok (chs a, char *v, char *w) {
       return 1;
     if (B.heads && aa.h >= B.heads) {
 	warn(_("%s of partition %s has impossible value for head: "
-	     "%d (should be in 0-%d)\n"), w, v, aa.h, B.heads-1);
+	     "%lu (should be in 0-%lu)\n"), w, v, aa.h, B.heads-1);
 	ret = 0;
     }
     if (B.sectors && (aa.s == 0 || aa.s > B.sectors)) {
 	warn(_("%s of partition %s has impossible value for sector: "
-	     "%d (should be in 1-%d)\n"), w, v, aa.s, B.sectors);
+	     "%lu (should be in 1-%lu)\n"), w, v, aa.s, B.sectors);
 	ret = 0;
     }
     if (B.cylinders && aa.c >= B.cylinders) {
 	warn(_("%s of partition %s has impossible value for cylinders: "
-	     "%d (should be in 0-%d)\n"), w, v, aa.c, B.cylinders-1);
+	     "%lu (should be in 0-%lu)\n"), w, v, aa.c, B.cylinders-1);
 	ret = 0;
     }
     return ret;
@@ -568,6 +567,7 @@ chs_ok (chs a, char *v, char *w) {
 #define LINUX_NATIVE	        0x83
 #define LINUX_EXTENDED		0x85
 #define BSD_PARTITION		0xa5
+#define NETBSD_PARTITION	0xa9
 
 /* List of partition types now in i386_sys_types.c */
 
@@ -599,7 +599,7 @@ is_extended(unsigned char type) {
 
 static int
 is_bsd(unsigned char type) {
-	return (type == BSD_PARTITION);
+	return (type == BSD_PARTITION || type == NETBSD_PARTITION);
 }
 
 /*
@@ -701,7 +701,7 @@ is_parent(struct part_desc *pp, struct part_desc *p) {
 }
 
 struct disk_desc {
-    struct part_desc partitions[128];
+    struct part_desc partitions[512];
     int partno;
 } oldp, newp;
 
@@ -867,24 +867,24 @@ out_partition_header(char *dev, int format, struct geometry G) {
 	  printf(_("Units = cylinders of %lu bytes, blocks of 1024 bytes"
 		 ", counting from %d\n\n"),
 		 G.cylindersize<<9, increment);
-	    printf(_("   Device Boot Start     End   #cyls   #blocks   Id  System\n"));
+	    printf(_("   Device Boot Start     End   #cyls    #blocks   Id  System\n"));
 	    break;
 	}
 	/* fall through */
       case F_SECTOR:
 	printf(_("Units = sectors of 512 bytes, counting from %d\n\n"),
 	       increment);
-	printf(_("   Device Boot    Start       End  #sectors  Id  System\n"));
+	printf(_("   Device Boot    Start       End   #sectors  Id  System\n"));
 	break;
       case F_BLOCK:
 	printf(_("Units = blocks of 1024 bytes, counting from %d\n\n"),
 	       increment);
-	printf(_("   Device Boot   Start       End   #blocks   Id  System\n"));
+	printf(_("   Device Boot   Start       End    #blocks   Id  System\n"));
 	break;
       case F_MEGABYTE:
 	printf(_("Units = megabytes of 1048576 bytes, blocks of 1024 bytes"
 	       ", counting from %d\n\n"), increment);
-	printf(_("   Device Boot Start   End     MB   #blocks   Id  System\n"));
+	printf(_("   Device Boot Start   End     MB    #blocks   Id  System\n"));
 	break;
     }
 }
@@ -916,13 +916,41 @@ out_roundup_size(int width, unsigned long n, unsigned long unit) {
     putchar(' ');
 }
 
-static int
-get_fdisk_geometry(struct part_desc *p) {
+static struct geometry
+get_fdisk_geometry_one(struct part_desc *p) {
+    struct geometry G;
+
     chs b = p->p.end_chs;
     longchs bb = chs_to_longchs(b);
-    F.heads = bb.h+1;
-    F.sectors = bb.s;
-    F.cylindersize = F.heads*F.sectors;
+    G.heads = bb.h+1;
+    G.sectors = bb.s;
+    G.cylindersize = G.heads*G.sectors;
+    G.cylinders = G.start = 0;
+    return G;
+}
+
+static int
+get_fdisk_geometry(struct disk_desc *z) {
+    struct part_desc *p;
+    int pno, agree;
+    struct geometry G0, G;
+
+    agree = 0;
+    G0.heads = G0.sectors = 0;
+    for (pno=0; pno < z->partno; pno++) {
+	p = &(z->partitions[pno]);
+	if (p->size != 0 && p->p.sys_type != 0) {
+	    G = get_fdisk_geometry_one(p);
+	    if (!G0.heads) {
+		G0 = G;
+		agree = 1;
+	    } else if (G.heads != G0.heads || G.sectors != G0.sectors) {
+		agree = 0;
+		break;
+	    }
+	}
+    }
+    F = (agree ? G0 : B);
     return (F.sectors != B.sectors || F.heads != B.heads);
 }
 
@@ -951,7 +979,7 @@ out_partition(char *dev, int format, struct part_desc *p,
 
     if (dump) {
 	printf(" start=%9lu", start);
-	printf(", size=%8lu", size);
+	printf(", size=%9lu", size);
 	if (p->ptype == DOS_TYPE) {
 	    printf(", Id=%2x", p->p.sys_type);
 	    if (p->p.bootable == 0x80)
@@ -974,7 +1002,7 @@ out_partition(char *dev, int format, struct part_desc *p,
 	    out_rounddown(6, start, G.cylindersize, increment);
 	    out_roundup(6, end, G.cylindersize, increment);
 	    out_roundup_size(6, size, G.cylindersize);
-	    out_rounddown(8, size, 2, 0);
+	    out_rounddown(9, size, 2, 0);
 	    break;
 	}
 	/* fall through */
@@ -982,7 +1010,7 @@ out_partition(char *dev, int format, struct part_desc *p,
       case F_SECTOR:
 	out_rounddown(9, start, 1, increment);
 	out_roundup(9, end, 1, increment);
-	out_rounddown(9, size, 1, 0);
+	out_rounddown(10, size, 1, 0);
 	break;
       case F_BLOCK:
 #if 0
@@ -991,17 +1019,17 @@ out_partition(char *dev, int format, struct part_desc *p,
 #endif
 	out_rounddown(8, start, 2, increment);
 	out_roundup(8, end, 2, increment);
-	out_rounddown(8, size, 2, 0);
+	out_rounddown(9, size, 2, 0);
 	break;
       case F_MEGABYTE:
 	out_rounddown(5, start, 2048, increment);
 	out_roundup(5, end, 2048, increment);
 	out_roundup_size(5, size, 2048);
-	out_rounddown(8, size, 2, 0);
+	out_rounddown(9, size, 2, 0);
 	break;
     }
     if (p->ptype == DOS_TYPE) {
-        printf(" %2x  %s\n",
+	printf(" %2x  %s\n",
 	   p->p.sys_type, sysname(p->p.sys_type));
     } else {
 	printf("\n");
@@ -1033,24 +1061,19 @@ out_partition(char *dev, int format, struct part_desc *p,
 
 static void
 out_partitions(char *dev, struct disk_desc *z) {
-    struct part_desc *p;
     int pno, format = 0;
 
     if (z->partno == 0)
 	printf(_("No partitions found\n"));
     else {
-	for (pno=0; pno < z->partno; pno++) {
-	    p = &(z->partitions[pno]);
-	    if (p->size != 0 && p->p.sys_type != 0) {
-		if (get_fdisk_geometry(p) && !dump)
-		    printf(
-		   _("Warning: The first partition looks like it was made\n"
-		     "  for C/H/S=*/%ld/%ld (instead of %ld/%ld/%ld).\n"
-		     "For this listing I'll assume that geometry.\n"),
-		   F.heads, F.sectors, B.cylinders, B.heads, B.sectors);
-		break;
-	    }
+	if (get_fdisk_geometry(z) && !dump) {
+	    printf(
+	   _("Warning: The partition table looks like it was made\n"
+	     "  for C/H/S=*/%ld/%ld (instead of %ld/%ld/%ld).\n"
+	     "For this listing I'll assume that geometry.\n"),
+	   F.heads, F.sectors, B.cylinders, B.heads, B.sectors);
 	}
+
 	out_partition_header(dev, format, F);
 	for(pno=0; pno < z->partno; pno++) {
 	    out_partition(dev, format, &(z->partitions[pno]), z, F);
@@ -1093,7 +1116,7 @@ partitions_ok(struct disk_desc *z) {
 
     /* Have at least 4 partitions been defined? */
     if (partno < 4) {
-         if (!partno)
+	 if (!partno)
 	      fatal(_("no partition table present.\n"));
 	 else
 	      fatal(_("strange, only %d partitions defined.\n"), partno);
@@ -1143,12 +1166,13 @@ partitions_ok(struct disk_desc *z) {
        table sectors disjoint? */
     for (p = partitions; p < partitions+partno; p++)
       if (p->size && !is_extended(p->p.sys_type))
-        for (q = partitions; q < partitions+partno; q++)
-          if (is_extended(q->p.sys_type))
-            if (p->start <= q->start && p->start + p->size > q->start) {
-                warn(_("Warning: partition %s contains part of "), PNO(p));
-		warn(_("the partition table (sector %lu),\n"), q->start);
-		warn(_("and will destroy it when filled\n"));
+	for (q = partitions; q < partitions+partno; q++)
+	  if (is_extended(q->p.sys_type))
+	    if (p->start <= q->start && p->start + p->size > q->start) {
+		warn(_("Warning: partition %s contains part of "
+		       "the partition table (sector %lu),\n"
+		       "and will destroy it when filled\n"),
+		     PNO(p), q->start);
 		return 0;
 	    }
 
@@ -1159,11 +1183,12 @@ partitions_ok(struct disk_desc *z) {
 	  if(p->start == 0) {
 	      warn(_("Warning: partition %s starts at sector 0\n"), PNO(p));
 	      return 0;
-          }
-          if (p->size && p->start + p->size > ds) {
-	      warn(_("Warning: partition %s extends past end of disk\n"), PNO(p));
+	  }
+	  if (p->size && p->start + p->size > ds) {
+	      warn(_("Warning: partition %s extends past end of disk\n"),
+		   PNO(p));
 	      return 0;
-          }
+	  }
       }
     }
 
@@ -1175,8 +1200,8 @@ partitions_ok(struct disk_desc *z) {
 	if (p->p.sys_type == EXTENDED_PARTITION)
 	  ect++;
       if (ect > 1 && !Linux) {
-	  warn(_("Among the primary partitions, at most one can be extended\n"));
-	  warn(_(" (although this is not a problem under Linux)\n"));
+	  warn(_("Among the primary partitions, at most one can be extended\n"
+		 " (although this is not a problem under Linux)\n"));
 	  return 0;
       }
     }
@@ -1222,12 +1247,12 @@ partitions_ok(struct disk_desc *z) {
 		   "This does not matter for LILO, but the DOS MBR will "
 		   "not boot this disk.\n"));
 	      break;
-          }
+	  }
 	  if (p - partitions >= 4) {
 	      warn(_("Warning: usually one can boot from primary partitions "
 		   "only\nLILO disregards the `bootable' flag.\n"));
 	      break;
-          }
+	  }
       }
       if (pno == -1 || pno >= 4)
 	warn(_("Warning: no primary partition is marked bootable (active)\n"
@@ -1284,13 +1309,15 @@ extended_partition(char *dev, int fd, struct part_desc *ep, struct disk_desc *z)
 	/* This is BAD */
 	if (DOS_extended) {
 	    here = start -= (start % B.cylindersize);
-	    printf(_("Warning: shifted start of the extd partition from %ld to %ld\n"),
+	    printf(_("Warning: shifted start of the extd partition "
+		     "from %ld to %ld\n"
+		     "(For listing purposes only. "
+		     "Do not change its contents.)\n"),
 		   ep->start, start);
-	    printf(_("(For listing purposes only. Do not change its contents.)\n"));
 	} else {
 	    printf(_("Warning: extended partition does not start at a "
-		     "cylinder boundary.\n"));
-	    printf(_("DOS and Linux will interpret the contents differently.\n"));
+		     "cylinder boundary.\n"
+		     "DOS and Linux will interpret the contents differently.\n"));
 	}
     }
 
@@ -1332,7 +1359,7 @@ extended_partition(char *dev, int fd, struct part_desc *ep, struct disk_desc *z)
 	    partitions[pno].ptype = DOS_TYPE;
 	    partitions[pno].p = p;
 	    pno++;
-        }
+	}
 	here = next;
     }
 
@@ -1341,19 +1368,19 @@ extended_partition(char *dev, int fd, struct part_desc *ep, struct disk_desc *z)
 
 #define BSD_DISKMAGIC   (0x82564557UL)
 #define BSD_MAXPARTITIONS       8
-#define BSD_FS_UNUSED           0
+#define BSD_FS_UNUSED	   0
 typedef unsigned char u8;
 typedef unsigned short u16;
 typedef unsigned int u32;
 struct bsd_disklabel {
 	u32	d_magic;
 	char	d_junk1[4];
-        char    d_typename[16];
-        char    d_packname[16];
+	char    d_typename[16];
+	char    d_packname[16];
 	char	d_junk2[92];
-        u32	d_magic2;
+	u32	d_magic2;
 	char	d_junk3[2];
-        u16	d_npartitions;          /* number of partitions in following */
+	u16	d_npartitions;          /* number of partitions in following */
 	char	d_junk4[8];
      struct  bsd_partition {         /* the partition table */
                 u32   p_size;         /* number of sectors in partition */
@@ -1542,7 +1569,8 @@ write_partitions(char *dev, int fd, struct disk_desc *z) {
 	if (!s) return 0;
 	s->to_be_written = 1;
 	copy_from_part(&(p->p), s->data + p->offset);
-	*(unsigned short *)(&(s->data[0x1fe])) = 0xaa55;
+	s->data[510] = 0x55;
+	s->data[511] = 0xaa;
     }
     if (save_sector_file) {
 	if (!save_sectors(dev, fd)) {
@@ -1701,13 +1729,23 @@ read_stdin(unsigned char **fields, unsigned char *line, int fieldssize, int line
 }
 
 /* read a number, use default if absent */
+/* a sign gives an offset from the default */
 static int
 get_ul(char *u, unsigned long *up, unsigned long def, int base) {
     char *nu;
+    int sign = 0;
+    unsigned long val;
 
+    if (*u == '+') {
+	sign = 1;
+	u++;
+    } else if (*u == '-') {
+	sign = -1;
+	u++;
+    }
     if (*u) {
 	errno = 0;
-	*up = strtoul(u, &nu, base);
+	val = strtoul(u, &nu, base);
 	if (errno == ERANGE) {
 	    printf(_("number too big\n"));
 	    return -1;
@@ -1716,6 +1754,11 @@ get_ul(char *u, unsigned long *up, unsigned long def, int base) {
 	    printf(_("trailing junk after number\n"));
 	    return -1;
 	}
+	if (sign == 1)
+		val = def + val;
+	else if (sign == -1)
+		val = def - val;
+	*up = val;
     } else
       *up = def;
     return 0;
@@ -2277,41 +2320,73 @@ static const struct option long_opts[] = {
     { NULL, 0, NULL, 0 }
 };
 
-/* default devices to list */
-static struct devd {
-    char *pref, *letters;
-} defdevs[] = {
-    { "hd", "abcdefgh" },
-    { "sd", "abcde" },
-    { "xd", "ab" },
-    { "ed", "abcd" }
-};
-
 static int
-is_ide_cdrom(char *device) {
-    /* No device was given explicitly, and we are trying some
-       likely things.  But opening /dev/hdc may produce errors like
+is_ide_cdrom_or_tape(char *device) {
+	FILE *procf;
+	char buf[100];
+	struct stat statbuf;
+	int is_ide = 0;
+
+	/* No device was given explicitly, and we are trying some
+	   likely things.  But opening /dev/hdc may produce errors like
            "hdc: tray open or drive not ready"
-       if it happens to be a CD-ROM drive. So try to be careful.
-       This only works since 2.1.73. */
+	   if it happens to be a CD-ROM drive. It even happens that
+	   the process hangs on the attempt to read a music CD.
+	   So try to be careful. This only works since 2.1.73. */
 
-    FILE *procf;
-    char buf[100];
-    struct stat statbuf;
+	if (strncmp("/dev/hd", device, 7))
+		return 0;
 
-    sprintf(buf, "/proc/ide/%s/media", device+5);
-    procf = fopen(buf, "r");
-    if (procf != NULL && fgets(buf, sizeof(buf), procf))
-	return  !strncmp(buf, "cdrom", 5);
+	snprintf(buf, sizeof(buf), "/proc/ide/%s/media", device+5);
+	procf = fopen(buf, "r");
+	if (procf != NULL && fgets(buf, sizeof(buf), procf))
+		is_ide = (!strncmp(buf, "cdrom", 5) ||
+			  !strncmp(buf, "tape", 4));
+	else
+		/* Now when this proc file does not exist, skip the
+		   device when it is read-only. */
+		if (stat(device, &statbuf) == 0)
+			is_ide = ((statbuf.st_mode & 0222) == 0);
 
-    /* Now when this proc file does not exist, skip the
-       device when it is read-only. */
-    if (stat(device, &statbuf) == 0)
-	return (statbuf.st_mode & 0222) == 0;
-
-    return 0;
+	if (procf)
+		fclose(procf);
+	return is_ide;
 }
 
+#define PROC_PARTITIONS	"/proc/partitions"
+static FILE *procf = NULL;
+
+static void
+openproc(void) {
+	procf = fopen(PROC_PARTITIONS, "r");
+	if (procf == NULL)
+		fprintf(stderr, _("cannot open %s\n"), PROC_PARTITIONS);
+}
+
+static char *
+nextproc(void) {
+	static char devname[120];
+	char line[100], ptname[100], *s;
+	int ma, mi, sz;
+
+	if (procf == NULL)
+		return NULL;
+	while (fgets(line, sizeof(line), procf) != NULL) {
+		if (sscanf (line, " %d %d %d %[^\n ]",
+			    &ma, &mi, &sz, ptname) != 4)
+			continue;
+		for(s = ptname; *s; s++);
+		if (isdigit(s[-1]))
+			continue;
+		snprintf(devname, sizeof(devname), "/dev/%s", ptname);
+		return devname;
+	}
+
+	fclose(procf);
+	procf = NULL;
+	return NULL;
+}	
+
 static void do_list(char *dev, int silent);
 static void do_size(char *dev, int silent);
 static void do_geom(char *dev, int silent);
@@ -2446,25 +2521,18 @@ main(int argc, char **argv) {
     }
 
     if (optind == argc && (opt_list || opt_out_geom || opt_size || verify)) {
-	struct devd *dp;
-	char *lp;
-	char device[10];
-
+	/* try all known devices */
 	total_size = 0;
-
-	for(dp = defdevs; dp-defdevs < SIZE(defdevs); dp++) {
-	    lp = dp->letters;
-	    while(*lp) {
-		sprintf(device, "/dev/%s%c", dp->pref, *lp++);
-		if (!strcmp(dp->pref, "hd") && is_ide_cdrom(device))
-		  continue;
-		if (opt_out_geom)
-		  do_geom(device, 1);
-		if (opt_size)
-		  do_size(device, 1);
-		if (opt_list || verify)
-		  do_list(device, 1);
-	    }
+	openproc();
+	while ((dev = nextproc()) != NULL) {
+	    if (!strncmp(dev, "hd", 2) && is_ide_cdrom_or_tape(dev))
+		continue;
+	    if (opt_out_geom)
+		do_geom(dev, 1);
+	    if (opt_size)
+		do_size(dev, 1);
+	    if (opt_list || verify)
+		do_list(dev, 1);
 	}
 
 	if (opt_size)
@@ -2541,7 +2609,10 @@ my_open (char *dev, int rw, int silent) {
     fd = open(dev, mode);
     if (fd < 0 && !silent) {
 	perror(dev);
-	fatal(_("cannot open %s %s\n"), dev, rw ? _("read-write") : _("for reading"));
+	if (rw)
+		fatal(_("cannot open %s read-write\n"), dev);
+	else
+		fatal(_("cannot open %s for reading\n"), dev);
     }
     return fd;
 }
@@ -2774,7 +2845,7 @@ do_change_id(char *dev, char *pnam, char *id) {
     }
     i = strtoul(id, NULL, 16);
     if (i > 255)
-      fatal(_("Bad Id %x\n"), i);
+      fatal(_("Bad Id %lx\n"), i);
     z->partitions[pno].p.sys_type = i;
 
     if(write_partitions(dev, fd, z))
