@@ -13,11 +13,7 @@
    - The program uses BSD command line options to be used
      in connection with e.g. 'rlogind' i.e. 'new login'.
    
-   - HP features left out:          logging of bad login attempts in /etc/btmp,
-                                    they are sent to syslog
-   
-				    password expiry
-   
+   - HP features left out:	    password expiry
 				    '*' as login shell, add it if you need it
 
    - BSD features left out:         quota checks
@@ -52,7 +48,7 @@
    From 1992 till now (1997) this code for Linux has been maintained at
    ftp.daimi.aau.dk:/pub/linux/poe/
 
-   1999-02-22 Arkadiusz Mikiewicz <misiek@misiek.eu.org>
+   1999-02-22 Arkadiusz Mikiewicz <misiek@pld.ORG.PL>
     - added Native Language Support
    Sun Mar 21 1999 - Arnaldo Carvalho de Melo <acme@conectiva.com.br>
     - fixed strerr(errno) in gettext calls
@@ -119,6 +115,7 @@
 #include "pathnames.h"
 #include "my_crypt.h"
 #include "login.h"
+#include "xstrncpy.h"
 #include "nls.h"
 
 #ifdef __linux__
@@ -145,8 +142,11 @@
        syslog(LOG_ERR,"%s",pam_strerror(pamh, retcode)); \
        pam_end(pamh, retcode); exit(1); \
    }
-#  define PAM_END { retcode = pam_close_session(pamh,0); \
-		    pam_end(pamh,retcode); }
+#  define PAM_END { \
+	pam_setcred(pamh, PAM_DELETE_CRED); \
+	retcode = pam_close_session(pamh,0); \
+	pam_end(pamh,retcode); \
+}
 #endif
 
 #ifndef __linux__
@@ -164,23 +164,6 @@
 #include "setproctitle.h"
 #endif
 
-/*
- * RedHat writes:
- * we've got a REAL HACK to avoid telling people that they have
- * mail because the imap server has left a turd in their inbox.
- * It works, but it sucks...
- * It turns out that the turd is always 523 bytes long, so we
- * just check for that size.
- */
-/*
- * If you want to turn this strange hack off, set
-	#define REDHAT_IGNORED_MAILSIZE 0
- * In case people complain, this may become a configuration option,
- * or perhaps this hack is thrown out again.
- * A better solution would be to check the contents of this file..
- */
-#define REDHAT_IGNORED_MAILSIZE	523
-
 #if 0
 /* from before we had a lastlog.h file in linux */
 struct  lastlog
@@ -190,13 +173,14 @@ struct  lastlog
 };
 #endif
 
-
+#ifndef USE_PAM
 static void getloginname (void);
+static void checknologin (void);
+static int rootterm (char *ttyn);
+#endif
 static void timedout (int);
 static void sigint (int);
-static int rootterm (char *ttyn);
 static void motd (void);
-static void checknologin (void);
 static void dolastlog (int quiet);
 
 #ifndef __linux__
@@ -238,24 +222,30 @@ int     timeout = 60;		/* used in cryptocard.c */
 #endif
 
 struct	passwd *pwd;		/* used in cryptocard.c */
-struct  hostent hostaddress;	/* used in checktty.c */
-char	term[64], *hostname, *username, *tty;
+#if USE_PAM
+static struct passwd pwdcopy;
+#endif
+char    hostaddress[4];		/* used in checktty.c */
+char	*hostname;		/* idem */
+static char	*username, *tty_name, *tty_number;
 static char	thishost[100];
 static int	failures = 1;
+static pid_t	pid;
 
 #ifndef __linux__
 struct	sgttyb sgttyb;
 struct	tchars tc = {
-    CINTR, CQUIT, CSTART, CSTOP, CEOT, CBRK
-  };
+	CINTR, CQUIT, CSTART, CSTOP, CEOT, CBRK
+};
 struct	ltchars ltc = {
-    CSUSP, CDSUSP, CRPRNT, CFLUSH, CWERASE, CLNEXT
-  };
+	CSUSP, CDSUSP, CRPRNT, CFLUSH, CWERASE, CLNEXT
+};
 #endif
 
-const char *months[] =
-{ "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug",
-    "Sep", "Oct", "Nov", "Dec" };
+const char *months[] = {
+	"Jan", "Feb", "Mar", "Apr", "May", "Jun",
+	"Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
+};
 
 /* Nice and simple code provided by Linus Torvalds 16-Feb-93 */
 /* Nonblocking stuff by Maciej W. Rozycki, macro@ds2.pg.gda.pl, 1999.
@@ -266,27 +256,33 @@ const char *months[] =
    leaving the decision to make a connection to getty (where it actually
    belongs). */
 static void
-opentty(const char * tty)
-{
-    int i;
-    int fd = open(tty, O_RDWR | O_NONBLOCK);
-    int flags = fcntl(fd, F_GETFL);
+opentty(const char * tty) {
+	int i, fd, flags;
 
-    flags &= ~O_NONBLOCK;
-    fcntl(fd, F_SETFL, flags);
+	fd = open(tty, O_RDWR | O_NONBLOCK);
+	if (fd == -1) {
+		syslog(LOG_ERR, _("FATAL: can't reopen tty: %s"),
+		       strerror(errno));
+		sleep(1);
+		exit(1);
+	}
+
+	flags = fcntl(fd, F_GETFL);
+	flags &= ~O_NONBLOCK;
+	fcntl(fd, F_SETFL, flags);
     
-    for (i = 0 ; i < fd ; i++)
-      close(i);
-    for (i = 0 ; i < 3 ; i++)
-      dup2(fd, i);
-    if (fd >= 3)
-      close(fd);
+	for (i = 0; i < fd; i++)
+		close(i);
+	for (i = 0; i < 3; i++)
+		if (fd != i)
+			dup2(fd, i);
+	if (fd >= 3)
+		close(fd);
 }
 
 /* true if the filedescriptor fd is a console tty, very Linux specific */
 static int
-consoletty(int fd)
-{
+consoletty(int fd) {
 #ifdef __linux__
     struct stat stb;
 
@@ -299,6 +295,47 @@ consoletty(int fd)
     return 0;
 }
 
+#if USE_PAM
+/*
+ * Log failed login attempts in _PATH_BTMP if that exists.
+ * Must be called only with username the name of an actual user.
+ * The most common login failure is to give password instead of username.
+ */
+#define	_PATH_BTMP	"/var/log/btmp"
+static void
+logbtmp(const char *line, const char *username, const char *hostname) {
+	struct utmp ut;
+
+	memset(&ut, 0, sizeof(ut));
+
+	strncpy(ut.ut_user, username ? username : "(unknown)",
+		sizeof(ut.ut_user));
+
+	strncpy(ut.ut_id, line + 3, sizeof(ut.ut_id));
+	xstrncpy(ut.ut_line, line, sizeof(ut.ut_line));
+
+#if defined(_HAVE_UT_TV)	    /* in <utmpbits.h> included by <utmp.h> */
+	gettimeofday(&ut.ut_tv, NULL);
+#else
+	{
+		time_t t;
+		time(&t);
+		ut.ut_time = t;	    /* ut_time is not always a time_t */
+	}
+#endif
+
+	ut.ut_type = LOGIN_PROCESS; /* XXX doesn't matter */
+	ut.ut_pid = pid;
+	if (hostname) {
+		xstrncpy(ut.ut_host, hostname, sizeof(ut.ut_host));
+		if (hostaddress[0])
+			memcpy(&ut.ut_addr, hostaddress, sizeof(ut.ut_addr));
+	}
+#ifdef HAVE_updwtmp		/* bad luck for ancient systems */
+	updwtmp(_PATH_BTMP, &ut);
+#endif
+}
+#endif	/* USE_PAM */
 
 int
 main(int argc, char **argv)
@@ -313,7 +350,6 @@ main(int argc, char **argv)
     char *domain, *ttyn;
     char tbuf[MAXPATHLEN + 2], tname[sizeof(_PATH_TTY) + 10];
     char *termenv;
-    char vcsn[20], vcsan[20];
     char * childArgv[10];
     char * buff;
     int childArgc = 0;
@@ -325,10 +361,15 @@ main(int argc, char **argv)
 #else
     char *salt, *pp;
 #endif
+#ifdef CHOWNVCS
+    char vcsn[20], vcsan[20];
+#endif
 #ifndef __linux__
     int ioctlval;
 #endif
 
+    pid = getpid();
+
     signal(SIGALRM, timedout);
     alarm((unsigned int)timeout);
     signal(SIGQUIT, SIG_IGN);
@@ -353,15 +394,14 @@ main(int argc, char **argv)
      *    host to login so that it may be placed in utmp and wtmp
      */
     gethostname(tbuf, sizeof(tbuf));
-    strncpy(thishost, tbuf, sizeof(thishost)-1);
-    thishost[sizeof(thishost)-1] = 0;
+    xstrncpy(thishost, tbuf, sizeof(thishost));
     domain = index(tbuf, '.');
     
-    username = tty = hostname = NULL;
+    username = tty_name = hostname = NULL;
     fflag = hflag = pflag = 0;
     passwd_req = 1;
 
-    while ((ch = getopt(argc, argv, "fh:p")) != EOF)
+    while ((ch = getopt(argc, argv, "fh:p")) != -1)
       switch (ch) {
 	case 'f':
 	  fflag = 1;
@@ -377,14 +417,16 @@ main(int argc, char **argv)
 	  if (domain && (p = index(optarg, '.')) &&
 	      strcasecmp(p, domain) == 0)
 	    *p = 0;
+
 	  hostname = strdup(optarg); 	/* strdup: Ambrose C. Li */
-	  { 
-	      struct hostent *he = gethostbyname(hostname);
-	      if (he) {
-		  memcpy(&hostaddress, he, sizeof(hostaddress));
-	      } else {
-		  memset(&hostaddress, 0, sizeof(hostaddress));
-	      }
+	  {
+		  struct hostent *he = gethostbyname(hostname);
+
+		  /* he points to static storage; copy the part we use */
+		  hostaddress[0] = 0;
+		  if (he && he->h_addr_list && he->h_addr_list[0])
+			  memcpy(hostaddress, he->h_addr_list[0],
+				 sizeof(hostaddress));
 	  }
 	  break;
 	  
@@ -424,8 +466,7 @@ main(int argc, char **argv)
     ioctl(0, TIOCSETP, &sgttyb);
     
     /*
-     * Be sure that we're in
-     * blocking mode!!!
+     * Be sure that we're in blocking mode!!!
      * This is really for HPUX
      */
     ioctlval = 0;
@@ -442,17 +483,28 @@ main(int argc, char **argv)
 	ttyn = tname;
     }
 
-    /* find names of Virtual Console devices, for later mode change */
-    {
-	char *p = ttyn;
-	/* find number of tty */
-	while (*p && !isdigit(*p)) p++;
+    if (strncmp(ttyn, "/dev/", 5) == 0)
+	tty_name = ttyn+5;
+    else
+	tty_name = ttyn;
 
-	strcpy(vcsn, "/dev/vcs"); strcat(vcsn, p);
-	strcpy(vcsan, "/dev/vcsa"); strcat(vcsan, p);
+    if (strncmp(ttyn, "/dev/tty", 8) == 0)
+	tty_number = ttyn+8;
+    else {
+	char *p = ttyn;
+	while (*p && !isdigit(*p)) p++;
+	tty_number = p;
     }
 
+#ifdef CHOWNVCS
+    /* find names of Virtual Console devices, for later mode change */
+    snprintf(vcsn, sizeof(vcsn), "/dev/vcs%s", tty_number);
+    snprintf(vcsan, sizeof(vcsan), "/dev/vcsa%s", tty_number);
+#endif
+
+    /* set pgid to pid */
     setpgrp();
+    /* this means that setsid() will fail */
     
     {
 	struct termios tt, ttt;
@@ -460,27 +512,24 @@ main(int argc, char **argv)
 	tcgetattr(0, &tt);
 	ttt = tt;
 	ttt.c_cflag &= ~HUPCL;
-	
-	if((chown(ttyn, 0, 0) == 0) && (chmod(ttyn, 0622) == 0)) {
-	    tcsetattr(0,TCSAFLUSH,&ttt);
-	    signal(SIGHUP, SIG_IGN); /* so vhangup() wont kill us */
-	    vhangup();
-	    signal(SIGHUP, SIG_DFL);
-	}
-	
-	setsid();
-	
-	/* re-open stdin,stdout,stderr after vhangup() closed them */
-	/* if it did, after 0.99.5 it doesn't! */
+
+	/* These can fail, e.g. with ttyn on a read-only filesystem */
+	chown(ttyn, 0, 0);
+	chmod(ttyn, TTY_MODE);
+
+	/* Kill processes left on this tty */
+	tcsetattr(0,TCSAFLUSH,&ttt);
+	signal(SIGHUP, SIG_IGN); /* so vhangup() wont kill us */
+	vhangup();
+	signal(SIGHUP, SIG_DFL);
+
+	/* open stdin,stdout,stderr to the tty */
 	opentty(ttyn);
+	
+	/* restore tty modes */
 	tcsetattr(0,TCSAFLUSH,&tt);
     }
 
-    if (strncmp(ttyn, "/dev/", 5) == 0)
-	tty = ttyn+5;
-    else
-	tty = ttyn;
-
     openlog("login", LOG_ODELAY, LOG_AUTHPRIV);
 
 #if 0
@@ -498,16 +547,17 @@ main(int argc, char **argv)
 
     retcode = pam_start("login",username, &conv, &pamh);
     if(retcode != PAM_SUCCESS) {
-	fprintf(stderr,_("login: PAM Failure, aborting: %s\n"),
-	pam_strerror(pamh, retcode));
-	syslog(LOG_ERR,_("Couldn't initialize PAM: %s"), pam_strerror(pamh, retcode));
+	fprintf(stderr, _("login: PAM Failure, aborting: %s\n"),
+		pam_strerror(pamh, retcode));
+	syslog(LOG_ERR, _("Couldn't initialize PAM: %s"),
+	       pam_strerror(pamh, retcode));
 	exit(99);
     }
     /* hostname & tty are either set to NULL or their correct values,
        depending on how much we know */
     retcode = pam_set_item(pamh, PAM_RHOST, hostname);
     PAM_FAIL_CHECK;
-    retcode = pam_set_item(pamh, PAM_TTY, tty);
+    retcode = pam_set_item(pamh, PAM_TTY, tty_name);
     PAM_FAIL_CHECK;
 
     /*
@@ -537,6 +587,11 @@ main(int argc, char **argv)
     if(passwd_req == 1) {
 	int failcount=0;
 
+	/* if we didn't get a user on the command line, set it to NULL */
+	pam_get_item(pamh,  PAM_USER, (const void **) &username);
+	if (!username)
+		pam_set_item(pamh, PAM_USER, NULL);
+
 	/* there may be better ways to deal with some of these
 	   conditions, but at least this way I don't think we'll
 	   be giving away information... */
@@ -550,8 +605,11 @@ main(int argc, char **argv)
 	       (retcode == PAM_CRED_INSUFFICIENT) ||
 	       (retcode == PAM_AUTHINFO_UNAVAIL))) {
 	    pam_get_item(pamh, PAM_USER, (const void **) &username);
+
 	    syslog(LOG_NOTICE,_("FAILED LOGIN %d FROM %s FOR %s, %s"),
-	    failcount, hostname, username, pam_strerror(pamh, retcode));
+		   failcount, hostname, username, pam_strerror(pamh, retcode));
+	    logbtmp(tty_name, username, hostname);
+
 	    fprintf(stderr,_("Login incorrect\n\n"));
 	    pam_set_item(pamh,PAM_USER,NULL);
 	    retcode = pam_authenticate(pamh, 0);
@@ -560,13 +618,14 @@ main(int argc, char **argv)
 	if (retcode != PAM_SUCCESS) {
 	    pam_get_item(pamh, PAM_USER, (const void **) &username);
 
-	    if (retcode == PAM_MAXTRIES) 
+	    if (retcode == PAM_MAXTRIES)
 		syslog(LOG_NOTICE,_("TOO MANY LOGIN TRIES (%d) FROM %s FOR "
 			"%s, %s"), failcount, hostname, username,
 			 pam_strerror(pamh, retcode));
 	    else
 		syslog(LOG_NOTICE,_("FAILED LOGIN SESSION FROM %s FOR %s, %s"),
 			hostname, username, pam_strerror(pamh, retcode));
+	    logbtmp(tty_name, username, hostname);
 
 	    fprintf(stderr,_("\nLogin incorrect\n"));
 	    pam_end(pamh, retcode);
@@ -582,20 +641,66 @@ main(int argc, char **argv)
 	PAM_FAIL_CHECK;
     }
 
-    /* Grab the user information out of the password file for future usage
-       First get the username that we are actually using, though.
-    */
+    /*
+     * Grab the user information out of the password file for future usage
+     * First get the username that we are actually using, though.
+     */
     retcode = pam_get_item(pamh, PAM_USER, (const void **) &username);
-    setpwent();
-    pwd = getpwnam(username);
-    if (pwd) initgroups(username, pwd->pw_gid);
-
-    retcode = pam_setcred(pamh, PAM_ESTABLISH_CRED);
     PAM_FAIL_CHECK;
 
+    if (!username || !*username) {
+	    fprintf(stderr, _("\nSession setup problem, abort.\n"));
+	    syslog(LOG_ERR, _("NULL user name in %s:%d. Abort."),
+		   __FUNCTION__, __LINE__);
+	    pam_end(pamh, PAM_SYSTEM_ERR);
+	    exit(1);
+    }
+    if (!(pwd = getpwnam(username))) {
+	    fprintf(stderr, _("\nSession setup problem, abort.\n"));
+	    syslog(LOG_ERR, _("Invalid user name \"%s\" in %s:%d. Abort."),
+		   username, __FUNCTION__, __LINE__);
+	    pam_end(pamh, PAM_SYSTEM_ERR);
+	    exit(1);
+    }
+
+    /*
+     * Create a copy of the pwd struct - otherwise it may get
+     * clobbered by PAM
+     */
+    memcpy(&pwdcopy, pwd, sizeof(*pwd));
+    pwd = &pwdcopy;
+    pwd->pw_name   = strdup(pwd->pw_name);
+    pwd->pw_passwd = strdup(pwd->pw_passwd);
+    pwd->pw_gecos  = strdup(pwd->pw_gecos);
+    pwd->pw_dir    = strdup(pwd->pw_dir);
+    pwd->pw_shell  = strdup(pwd->pw_shell);
+    if (!pwd->pw_name || !pwd->pw_passwd || !pwd->pw_gecos ||
+	!pwd->pw_dir || !pwd->pw_shell) {
+	    fprintf(stderr, _("login: Out of memory\n"));
+	    syslog(LOG_ERR, "Out of memory");
+	    pam_end(pamh, PAM_SYSTEM_ERR);
+	    exit(1);
+    }
+    username = pwd->pw_name;
+
+    /*
+     * Initialize the supplementary group list.
+     * This should be done before pam_setcred because
+     * the PAM modules might add groups during pam_setcred.
+     */
+    if (initgroups(username, pwd->pw_gid) < 0) {
+	    syslog(LOG_ERR, "initgroups: %m");
+	    fprintf(stderr, _("\nSession setup problem, abort.\n"));
+	    pam_end(pamh, PAM_SYSTEM_ERR);
+	    exit(1);
+    }
+
     retcode = pam_open_session(pamh, 0);
     PAM_FAIL_CHECK;
 
+    retcode = pam_setcred(pamh, PAM_ESTABLISH_CRED);
+    PAM_FAIL_CHECK;
+
 #else /* ! USE_PAM */
 
     for (cnt = 0;; ask = 1) {
@@ -636,7 +741,7 @@ main(int argc, char **argv)
 	
 	if (pwd) {
 	    initgroups(username, pwd->pw_gid);
-	    checktty(username, tty, pwd); /* in checktty.c */
+	    checktty(username, tty_name, pwd); /* in checktty.c */
 	}
 	
 	/* if user not super-user, check for disabled logins */
@@ -658,7 +763,7 @@ main(int argc, char **argv)
 	 * If trying to log in as root, but with insecure terminal,
 	 * refuse the login attempt.
 	 */
-	if (pwd && pwd->pw_uid == 0 && !rootterm(tty)) {
+	if (pwd && pwd->pw_uid == 0 && !rootterm(tty_name)) {
 	    fprintf(stderr,
 		    _("%s login refused on this terminal.\n"),
 		    pwd->pw_name);
@@ -666,11 +771,11 @@ main(int argc, char **argv)
 	    if (hostname)
 	      syslog(LOG_NOTICE,
 		     _("LOGIN %s REFUSED FROM %s ON TTY %s"),
-		     pwd->pw_name, hostname, tty);
+		     pwd->pw_name, hostname, tty_name);
 	    else
 	      syslog(LOG_NOTICE,
 		     _("LOGIN %s REFUSED ON TTY %s"),
-		     pwd->pw_name, tty);
+		     pwd->pw_name, tty_name);
 	    continue;
 	}
 
@@ -753,7 +858,7 @@ main(int argc, char **argv)
 	  default:
 	    perror("quota (Q_SETUID)");
 	}
-	sleepexit(0);
+	sleepexit(0);		/* %% */
     }
 #endif
     
@@ -791,76 +896,15 @@ main(int argc, char **argv)
 	}
     }
     
-#ifndef __linux__
-#  ifdef KERBEROS
-    if (notickets && !quietlog)
-      printf(_("Warning: no Kerberos tickets issued\n"));
-#  endif
-    
-#  ifndef USE_PAM			/* PAM does all of this for us */
-#    define TWOWEEKS	(14*24*60*60)
-    if (pwd->pw_change || pwd->pw_expire) {
-	struct timeval tp;
-
-	gettimeofday(&tp, (struct timezone *)NULL);
-
-	if (pwd->pw_change) {
-	    if (tp.tv_sec >= pwd->pw_change) {
-		printf(_("Sorry -- your password has expired.\n"));
-		sleepexit(1);
-	    }
-	    else if (tp.tv_sec - pwd->pw_change < TWOWEEKS && !quietlog) {
-		struct tm *ttp;
-		ttp = localtime(&pwd->pw_change);
-		printf(_("Warning: your password expires on %s %d, %d\n"),
-		       months[ttp->tm_mon], ttp->tm_mday,
-		       TM_YEAR_BASE + ttp->tm_year);
-	    }
-	}
-
-	if (pwd->pw_expire) {
-	    if (tp.tv_sec >= pwd->pw_expire) {
-		printf(_("Sorry -- your account has expired.\n"));
-		sleepexit(1);
-	    }
-	    else if (tp.tv_sec - pwd->pw_expire < TWOWEEKS && !quietlog) {
-		struct tm *ttp;
-		ttp = localtime(&pwd->pw_expire);
-		printf(_("Warning: your account expires on %s %d, %d\n"),
-		       months[ttp->tm_mon], ttp->tm_mday,
-		       TM_YEAR_BASE + ttp->tm_year);
-	    }
-	}
-    }
-#  endif /* !USE_PAM */
-    
-    /* nothing else left to fail -- really log in */
-    {
-	struct utmp utmp;
-	
-	memset((char *)&utmp, 0, sizeof(utmp));
-	time(&utmp.ut_time);
-	strncpy(utmp.ut_name, username, sizeof(utmp.ut_name));
-	/* ut_name may legally be non-null-terminated */
-	if (hostname) {
-	    strncpy(utmp.ut_host, hostname, sizeof(utmp.ut_host));
-	    utmp.ut_host[sizeof(utmp.ut_host)-1] = 0;
-	}
-	strncpy(utmp.ut_line, tty, sizeof(utmp.ut_line));
-	utmp.ut_line[sizeof(utmp.ut_line)-1] = 0;
-	login(&utmp);
-    }
-#else /* __linux__ defined */
     /* for linux, write entries in utmp and wtmp */
     {
 	struct utmp ut;
 	struct utmp *utp;
-	pid_t mypid = getpid();
 	
 	utmpname(_PATH_UTMP);
 	setutent();
 
-	/* Find mypid in utmp.
+	/* Find pid in utmp.
 login sometimes overwrites the runlevel entry in /var/run/utmp,
 confusing sysvinit. I added a test for the entry type, and the problem
 was gone. (In a runlevel entry, st_pid is not really a pid but some number
@@ -868,7 +912,7 @@ calculated from the previous and current runlevel).
 Michael Riepe <michael@stud.uni-hannover.de>
 	*/
 	while ((utp = getutent()))
-		if (utp->ut_pid == mypid
+		if (utp->ut_pid == pid
 		    && utp->ut_type >= INIT_PROCESS
 		    && utp->ut_type <= DEAD_PROCESS)
 			break;
@@ -878,8 +922,8 @@ Michael Riepe <michael@stud.uni-hannover.de>
 	if (utp == NULL) {
 	     setutent();
 	     ut.ut_type = LOGIN_PROCESS;
-	     strncpy(ut.ut_id, ttyn + 8, sizeof(ut.ut_id));
-	     strncpy(ut.ut_line, ttyn + 5, sizeof(ut.ut_line));
+	     strncpy(ut.ut_id, tty_number, sizeof(ut.ut_id));
+	     strncpy(ut.ut_line, tty_name, sizeof(ut.ut_line));
 	     utp = getutid(&ut);
 	}
 	
@@ -891,35 +935,32 @@ Michael Riepe <michael@stud.uni-hannover.de>
 	}
 	
 	if (ut.ut_id[0] == 0)
-	  strncpy(ut.ut_id, ttyn + 8, sizeof(ut.ut_id));
+	  strncpy(ut.ut_id, tty_number, sizeof(ut.ut_id));
 	
 	strncpy(ut.ut_user, username, sizeof(ut.ut_user));
-	strncpy(ut.ut_line, ttyn + 5, sizeof(ut.ut_line));
-	ut.ut_line[sizeof(ut.ut_line)-1] = 0;
+	xstrncpy(ut.ut_line, tty_name, sizeof(ut.ut_line));
 #ifdef _HAVE_UT_TV		/* in <utmpbits.h> included by <utmp.h> */
 	gettimeofday(&ut.ut_tv, NULL);
 #else
 	{
-	  time_t t;
-	  time(&t);
-	  ut.ut_time = t;	/* ut_time is not always a time_t */
+	    time_t t;
+	    time(&t);
+	    ut.ut_time = t;	/* ut_time is not always a time_t */
 				/* glibc2 #defines it as ut_tv.tv_sec */
 	}
 #endif
 	ut.ut_type = USER_PROCESS;
-	ut.ut_pid = mypid;
+	ut.ut_pid = pid;
 	if (hostname) {
-	    strncpy(ut.ut_host, hostname, sizeof(ut.ut_host));
-	    ut.ut_host[sizeof(ut.ut_host)-1] = 0;
-	    if (hostaddress.h_addr_list)
-	      memcpy(&ut.ut_addr, hostaddress.h_addr_list[0],
-		     sizeof(ut.ut_addr));
+		xstrncpy(ut.ut_host, hostname, sizeof(ut.ut_host));
+		if (hostaddress[0])
+			memcpy(&ut.ut_addr, hostaddress, sizeof(ut.ut_addr));
 	}
 	
 	pututline(&ut);
 	endutent();
 
-#if __GLIBC__ > 2 || (__GLIBC__ == 2 && __GLIBC_MINOR__ >= 1)
+#ifdef HAVE_updwtmp
 	updwtmp(_PATH_WTMP, &ut);
 #else
 #if 0
@@ -949,23 +990,16 @@ Michael Riepe <michael@stud.uni-hannover.de>
 	    }
 	}
 #endif
-#endif /* __GLIBC__ */
+#endif
     }
-#endif /* __linux__ */
     
     dolastlog(quietlog);
     
-#ifndef __linux__
-    if (!hflag) {					/* XXX */
-	static struct winsize win = { 0, 0, 0, 0 };
-	
-	ioctl(0, TIOCSWINSZ, &win);
-    }
-#endif
     chown(ttyn, pwd->pw_uid,
 	  (gr = getgrnam(TTYGRPNAME)) ? gr->gr_gid : pwd->pw_gid);
     chmod(ttyn, TTY_MODE);
 
+#ifdef CHOWNVCS
     /* if tty is one of the VC's then change owner and mode of the 
        special /dev/vcs devices as well */
     if (consoletty(0)) {
@@ -974,6 +1008,7 @@ Michael Riepe <michael@stud.uni-hannover.de>
 	chmod(vcsn, TTY_MODE);
 	chmod(vcsan, TTY_MODE);
     }
+#endif
 
     setgid(pwd->pw_gid);
     
@@ -983,13 +1018,6 @@ Michael Riepe <michael@stud.uni-hannover.de>
     
     if (*pwd->pw_shell == '\0')
       pwd->pw_shell = _PATH_BSHELL;
-#ifndef __linux__
-    /* turn on new line discipline for the csh */
-    else if (!strcmp(pwd->pw_shell, _PATH_CSHELL)) {
-	ioctlval = NTTYDISC;
-	ioctl(0, TIOCSETD, &ioctlval);
-    }
-#endif
     
     /* preserve TERM even without -p flag */
     {
@@ -1006,17 +1034,6 @@ Michael Riepe <michael@stud.uni-hannover.de>
 	  memset(environ, 0, sizeof(char*));
       }
     
-#ifndef __linux__
-    setenv("HOME", pwd->pw_dir, 1);
-    setenv("SHELL", pwd->pw_shell, 1);
-    if (term[0] == '\0') {
-	strncpy(term, stypeof(tty), sizeof(term));
-	term[sizeof(term)-1] = 0;
-    }
-    setenv("TERM", term, 0);
-    setenv("USER", pwd->pw_name, 1);
-    setenv("PATH", _PATH_DEFPATH, 0);
-#else
     setenv("HOME", pwd->pw_dir, 0);      /* legal to override */
     if(pwd->pw_uid)
       setenv("PATH", _PATH_DEFPATH, 1);
@@ -1040,19 +1057,16 @@ Michael Riepe <michael@stud.uni-hannover.de>
        HP-UX 6.5 does it. We'll not allow modifying it.
        */
     setenv("LOGNAME", pwd->pw_name, 1);
-#endif
 
 #ifdef USE_PAM
     {
 	int i;
-	const char * const * env;
-
-	env = (const char * const *)pam_getenvlist(pamh);
+	char ** env = pam_getenvlist(pamh);
 
 	if (env != NULL) {
-	    for (i=0; env[i++]; ) {
-		putenv(env[i-1]);
-		/* D(("env[%d] = %s", i-1,env[i-1])); */
+	    for (i=0; env[i]; i++) {
+		putenv(env[i]);
+		/* D(("env[%d] = %s", i,env[i])); */
 	    }
 	}
     }
@@ -1062,8 +1076,8 @@ Michael Riepe <michael@stud.uni-hannover.de>
     setproctitle("login", username);
 #endif
     
-    if (tty[sizeof("tty")-1] == 'S')
-      syslog(LOG_INFO, _("DIALUP AT %s BY %s"), tty, pwd->pw_name);
+    if (!strncmp(tty_name, "ttyS", 4))
+      syslog(LOG_INFO, _("DIALUP AT %s BY %s"), tty_name, pwd->pw_name);
     
     /* allow tracking of good logins.
        -steve philp (sphilp@mail.alliance.net) */
@@ -1071,15 +1085,15 @@ Michael Riepe <michael@stud.uni-hannover.de>
     if (pwd->pw_uid == 0) {
 	if (hostname)
 	  syslog(LOG_NOTICE, _("ROOT LOGIN ON %s FROM %s"),
-		 tty, hostname);
+		 tty_name, hostname);
 	else
-	  syslog(LOG_NOTICE, _("ROOT LOGIN ON %s"), tty);
+	  syslog(LOG_NOTICE, _("ROOT LOGIN ON %s"), tty_name);
     } else {
 	if (hostname) 
-	  syslog(LOG_INFO, _("LOGIN ON %s BY %s FROM %s"), tty, 
+	  syslog(LOG_INFO, _("LOGIN ON %s BY %s FROM %s"), tty_name, 
 		 pwd->pw_name, hostname);
 	else 
-	  syslog(LOG_INFO, _("LOGIN ON %s BY %s"), tty, 
+	  syslog(LOG_INFO, _("LOGIN ON %s BY %s"), tty_name, 
 		 pwd->pw_name);
     }
     
@@ -1089,37 +1103,66 @@ Michael Riepe <michael@stud.uni-hannover.de>
 	
 	motd();
 	mail = getenv("MAIL");
-	if (mail && stat(mail, &st) == 0 && st.st_size != 0
-		 && st.st_size != REDHAT_IGNORED_MAILSIZE) {
-		printf(_("You have %smail.\n"),
-		       (st.st_mtime > st.st_atime) ? _("new ") : "");
+	if (mail && stat(mail, &st) == 0 && st.st_size != 0) {
+		if (st.st_mtime > st.st_atime)
+			printf(_("You have new mail.\n"));
+		else
+			printf(_("You have mail.\n"));
 	}
     }
     
     signal(SIGALRM, SIG_DFL);
     signal(SIGQUIT, SIG_DFL);
     signal(SIGTSTP, SIG_IGN);
-    signal(SIGHUP, SIG_DFL);
 
 #ifdef USE_PAM
-    /* We must fork before setuid() because we need to call
+    /*
+     * We must fork before setuid() because we need to call
      * pam_close_session() as root.
      */
-    signal(SIGINT, SIG_IGN);
+    
     childPid = fork();
     if (childPid < 0) {
        int errsv = errno;
        /* error in fork() */
-       fprintf(stderr,_("login: failure forking: %s"), strerror(errsv));
+       fprintf(stderr, _("login: failure forking: %s"), strerror(errsv));
        PAM_END;
        exit(0);
-    } else if (childPid) {
+    }
+
+    if (childPid) {
        /* parent - wait for child to finish, then cleanup session */
+       signal(SIGHUP, SIG_IGN);
+       signal(SIGINT, SIG_IGN);
+       signal(SIGQUIT, SIG_IGN);
+       signal(SIGTSTP, SIG_IGN);
+       signal(SIGTTIN, SIG_IGN);
+       signal(SIGTTOU, SIG_IGN);
+
        wait(NULL);
        PAM_END;
        exit(0);
     }
+
     /* child */
+    /*
+     * Problem: if the user's shell is a shell like ash that doesnt do
+     * setsid() or setpgrp(), then a ctrl-\, sending SIGQUIT to every
+     * process in the pgrp, will kill us.
+     */
+
+    /* start new session */
+    setsid();
+
+    /* make sure we have a controlling tty */
+    opentty(ttyn);
+    openlog("login", LOG_ODELAY, LOG_AUTHPRIV);	/* reopen */
+
+    /*
+     * TIOCSCTTY: steal tty from other process group.
+     */
+    if (ioctl(0, TIOCSCTTY, 1))
+	    syslog(LOG_ERR, _("TIOCSCTTY failed: %m"));
 #endif
     signal(SIGINT, SIG_DFL);
     
@@ -1155,10 +1198,9 @@ Michael Riepe <michael@stud.uni-hannover.de>
 	childArgv[childArgc++] = buff;
     } else {
 	tbuf[0] = '-';
-	strncpy(tbuf + 1, ((p = rindex(pwd->pw_shell, '/')) ?
+	xstrncpy(tbuf + 1, ((p = rindex(pwd->pw_shell, '/')) ?
 			   p + 1 : pwd->pw_shell),
 		sizeof(tbuf)-1);
-	tbuf[sizeof(tbuf)-1] = 0;
 	
 	childArgv[childArgc++] = pwd->pw_shell;
 	childArgv[childArgc++] = tbuf;
@@ -1179,7 +1221,8 @@ Michael Riepe <michael@stud.uni-hannover.de>
     exit(0);
 }
 
-void
+#ifndef USE_PAM
+static void
 getloginname(void) {
     int ch, cnt, cnt2;
     char *p;
@@ -1223,6 +1266,7 @@ getloginname(void) {
 	}
     }
 }
+#endif
 
 void
 timedout(int sig) {
@@ -1234,16 +1278,17 @@ timedout(int sig) {
     ioctl(0, TCGETA, &ti);
     ti.c_lflag |= ECHO;
     ioctl(0, TCSETA, &ti);
-    exit(0);
+    exit(0);			/* %% */
 }
 
+#ifndef USE_PAM
 int
 rootterm(char * ttyn)
 #ifndef __linux__
 {
     struct ttyent *t;
     
-    return((t = getttynam(ttyn)) && t->ty_status&TTY_SECURE);
+    return((t = getttynam(ttyn)) && (t->ty_status&TTY_SECURE));
 }
 #else
 { 
@@ -1273,7 +1318,8 @@ rootterm(char * ttyn)
   	}
     }
 }
-#endif
+#endif /* !__linux__ */
+#endif /* !USE_PAM */
 
 jmp_buf motdinterrupt;
 
@@ -1301,12 +1347,13 @@ sigint(int sig) {
 #ifndef USE_PAM				/* PAM takes care of this */
 void
 checknologin(void) {
-    register int fd, nchars;
+    int fd, nchars;
     char tbuf[8192];
     
     if ((fd = open(_PATH_NOLOGIN, O_RDONLY, 0)) >= 0) {
 	while ((nchars = read(fd, tbuf, sizeof(tbuf))) > 0)
 	  write(fileno(stdout), tbuf, nchars);
+	close(fd);
 	sleepexit(0);
     }
 }
@@ -1336,34 +1383,31 @@ dolastlog(int quiet) {
 	}
 	memset((char *)&ll, 0, sizeof(ll));
 	time(&ll.ll_time);
-	strncpy(ll.ll_line, tty, sizeof(ll.ll_line));
-	ll.ll_line[sizeof(ll.ll_line)-1] = 0;
-	if (hostname) {
-	    strncpy(ll.ll_host, hostname, sizeof(ll.ll_host));
-	    ll.ll_host[sizeof(ll.ll_host)-1] = 0;
-	}
+	xstrncpy(ll.ll_line, tty_name, sizeof(ll.ll_line));
+	if (hostname)
+	    xstrncpy(ll.ll_host, hostname, sizeof(ll.ll_host));
+
 	write(fd, (char *)&ll, sizeof(ll));
 	close(fd);
     }
 }
 
 void
-badlogin(const char *name)
-{
+badlogin(const char *name) {
     if (failures == 1) {
 	if (hostname)
 	  syslog(LOG_NOTICE, _("LOGIN FAILURE FROM %s, %s"),
 		 hostname, name);
 	else
 	  syslog(LOG_NOTICE, _("LOGIN FAILURE ON %s, %s"),
-		 tty, name);
+		 tty_name, name);
     } else {
 	if (hostname)
 	  syslog(LOG_NOTICE, _("%d LOGIN FAILURES FROM %s, %s"),
 		 failures, hostname, name);
 	else
 	  syslog(LOG_NOTICE, _("%d LOGIN FAILURES ON %s, %s"),
-		 failures, tty, name);
+		 failures, tty_name, name);
     }
 }
 
@@ -1379,7 +1423,7 @@ stypeof(char *ttyid) {
 }
 #endif 
 
-/* should not be called from PAM code... Why? */
+/* Should not be called from PAM code... */
 void
 sleepexit(int eval) {
     sleep(SLEEP_EXIT_TIMEOUT);
