@@ -28,7 +28,7 @@
  * Changes by Richard Gooch <rgooch@atnf.csiro.au> (butchered by aeb)
  * introducing shutdown.conf.
  *
- * 1999-02-22 Arkadiusz Mikiewicz <misiek@misiek.eu.org>
+ * 1999-02-22 Arkadiusz Mikiewicz <misiek@pld.ORG.PL>
  * - added Native Language Support
  *
  * 2000-03-02 Richard Gooch <rgooch@atnf.csiro.au>
@@ -71,6 +71,7 @@
 #include <sys/utsname.h>
 #include "linux_reboot.h"
 #include "pathnames.h"
+#include "xstrncpy.h"
 #include "nls.h"
 
 static void usage(void), int_handler(int), write_user(struct utmp *);
@@ -249,8 +250,7 @@ main(int argc, char *argv[])
 				exit(1);
 			}
 		} else {
-			strncpy(message, argv[c], sizeof(message));
-			message[sizeof(message)-1] = '\0';
+			xstrncpy(message, argv[c], sizeof(message));
 			opt_msgset = 1;
 		}
 	}
@@ -463,6 +463,9 @@ write_user(struct utmp *ut)
 	char msg[100];
 
 	minutes = timeout / 60;
+	hours = minutes / 60;
+	minutes %= 60;
+
 	(void) strncat(term, ut->ut_line, sizeof(ut->ut_line));
 
 	/* try not to get stuck on a mangled ut_line entry... */
@@ -474,16 +477,20 @@ write_user(struct utmp *ut)
 	WR(msg);
 	WRCRLF;
 
-	if(minutes == 0) {
-	    sprintf(msg, _("System going down IMMEDIATELY!\n"));
-	} else if(minutes > 60) {
-	    hours = minutes / 60;
-	    sprintf(msg, _("System going down in %d hour%s %d minutes"),
-		    hours, hours == 1 ? "" : _("s"), minutes - 60*hours);
-	} else {
-	    sprintf(msg, _("System going down in %d minute%s\n"),
-		    minutes, minutes == 1 ? "" : _("s"));
-	}
+	if (hours > 1)
+		sprintf(msg, _("System going down in %d hours %d minutes"),
+			hours, minutes);
+	else if (hours == 1)
+		sprintf(msg, _("System going down in 1 hour %d minutes"),
+			minutes);
+	else if (minutes > 1)
+		sprintf(msg, _("System going down in %d minutes\n"),
+			minutes);
+	else if (minutes == 1)
+		sprintf(msg, _("System going down in 1 minute\n"));
+	else
+		sprintf(msg, _("System going down IMMEDIATELY!\n"));
+
 	WR(msg);
 	WRCRLF;
 
